import argparse
import logging
from typing import List

import torch

from .. import headmeta, metric

LOG = logging.getLogger(__name__)


class DataModule:
    """
    Base class to extend OpenPifPaf with custom data.

    This class gives you all the handles to train OpenPifPaf on a new dataset.
    Create a new class that inherits from this to handle a new datasets.


    1. Define the PifPaf heads you would like to train. \
    For example, \
    CIF (Composite Intensity Fields) to detect keypoints, and \
    CAF (Composite Association Fields) to associate joints \

    2. Add class variables, such as annotations, training/validation image paths.

    """

    #: Data loader batch size.
    batch_size = 1

    #: Data loader number of workers.
    _loader_workers = None

    #: A list of head metas for this dataset.
    #: Set as instance variable (not class variable) in derived classes
    #: so that different instances of head metas are created for different
    #: instances of the data module. Head metas contain the base stride which
    #: might be different for different data module instances.
    #: When loading a checkpoint, entries in this list will be matched by
    #: name and dataset to entries in the checkpoint and overwritten here.
    head_metas: List[headmeta.Base] = None

    @classmethod
    def set_loader_workers(cls, value):
        cls._loader_workers = value

    @property
    def loader_workers(self):
        if self._loader_workers is not None:
            return self._loader_workers

        # Do not propose more than 16 loaders. More loaders use more
        # shared memory. When shared memory is exceeded, all jobs
        # on that machine crash.
        return min(16, self.batch_size)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        r"""
        Command line interface (CLI) to extend argument parser for your custom dataset.

        Make sure to use unique CLI arguments for your dataset.
        For clarity, we suggest to start every CLI argument with the name of your new dataset,
        i.e. \-\-<dataset_name>-train-annotations.

        All PifPaf commands will still work.
        E.g. to load a model, there is no need to implement the command \-\-checkpoint
        """

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""

    def metrics(self) -> List[metric.Base]:
        """Define a list of metrics to be used for eval."""
        raise NotImplementedError

    def train_loader(self) -> torch.utils.data.DataLoader:
        """
        Loader of the training dataset.

        A Coco Data loader is already available, or a custom one can be created and called here.
        To modify preprocessing steps of your images (for example scaling image during training):

        1. chain them using torchvision.transforms.Compose(transforms)
        2. pass them to the preprocessing argument of the dataloader"""
        raise NotImplementedError

    def val_loader(self) -> torch.utils.data.DataLoader:
        """
        Loader of the validation dataset.

        The augmentation and preprocessing should be the same as for train_loader.
        The only difference is the set of data. This allows to inspect the
        train/val curves for overfitting.

        As in the train_loader, the annotations should be encoded fields
        so that the loss function can be computed.
        """
        raise NotImplementedError

    def eval_loader(self) -> torch.utils.data.DataLoader:
        """
        Loader of the evaluation dataset.

        For local runs, it is common that the validation dataset is also the
        evaluation dataset. This is then changed to test datasets (without
        ground truth) to produce predictions for submissions to a competition
        server that holds the private ground truth.

        This loader shouldn't have any data augmentation. The images should be
        as close as possible to the real application.
        The annotations should be the ground truth annotations similarly to
        what the output of the decoder is expected to be.
        """
        raise NotImplementedError

    @staticmethod
    def distributed_sampler(loader: torch.utils.data.DataLoader):
        LOG.info('Replacing sampler of %s with DistributedSampler.', loader)
        distributed_sampler = torch.utils.data.DistributedSampler(
            loader.dataset, shuffle=True, drop_last=True)

        return torch.utils.data.DataLoader(
            loader.dataset,
            batch_size=loader.batch_size,
            drop_last=True,
            shuffle=False,
            sampler=distributed_sampler,
            pin_memory=loader.pin_memory,
            num_workers=loader.num_workers,
            collate_fn=loader.collate_fn,
        )



"""
Convert txt files of ApolloCar3D into json file with COCO format
"""

import glob
import os
import time
from shutil import copyfile
import json
import argparse

import numpy as np
from PIL import Image

# Packages for data processing, crowd annotations and histograms
try:
    import matplotlib.pyplot as plt  # pylint: disable=import-error
except ModuleNotFoundError as err:
    if err.name != 'matplotlib':
        raise err
    plt = None
try:
    import cv2  # pylint: disable=import-error
except ModuleNotFoundError as err:
    if err.name != 'cv2':
        raise err
    cv2 = None  # pylint: disable=invalid-name

from .constants import CAR_KEYPOINTS_24, CAR_SKELETON_24,\
    CAR_KEYPOINTS_66, CAR_SKELETON_66, KPS_MAPPING
from .transforms import skeleton_mapping


def cli():
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--dir_data', default='data-apollocar3d/train',
                        help='dataset directory')
    parser.add_argument('--dir_out', default='data-apollocar3d',
                        help='where to save annotations and files')
    parser.add_argument('--sample', action='store_true',
                        help='Whether to only process the first 50 images')
    parser.add_argument('--single_sample', action='store_true',
                        help='Whether to only process the first image')
    parser.add_argument('--split_images', action='store_true',
                        help='Whether to copy images into train val split folder')
    parser.add_argument('--histogram', action='store_true',
                        help='Whether to show keypoints histogram')
    args = parser.parse_args()
    return args


class ApolloToCoco:

    # Prepare json format
    map_sk = skeleton_mapping(KPS_MAPPING)

    sample = False
    single_sample = False
    split_images = False
    histogram = False

    def __init__(self, dir_dataset, dir_out):
        """
        :param dir_dataset: Original dataset directory
        :param dir_out: Processed dataset directory
        """

        assert os.path.isdir(dir_dataset), 'dataset directory not found'
        self.dir_dataset = dir_dataset
        self.dir_mask = os.path.join(dir_dataset, 'ignore_mask')
        assert os.path.isdir(self.dir_mask), 'crowd annotations not found: ' + self.dir_mask

        self.dir_out_im = os.path.join(dir_out, 'images')
        self.dir_out_ann = os.path.join(dir_out, 'annotations')
        os.makedirs(self.dir_out_im, exist_ok=True)
        os.makedirs(self.dir_out_ann, exist_ok=True)

        self.json_file_24 = {}
        self.json_file_66 = {}

        # Load train val split
        path_train = os.path.join(self.dir_dataset, 'split', 'train-list.txt')
        path_val = os.path.join(self.dir_dataset, 'split', 'validation-list.txt', )
        self.splits = {}
        for name, path in zip(('train', 'val'), (path_train, path_val)):
            with open(path, "r") as ff:
                lines = ff.readlines()
            self.splits[name] = [os.path.join(self.dir_dataset, 'images', line.strip())
                                 for line in lines]
            assert self.splits[name], "specified path is empty"

    def process(self):
        """Parse and process the txt dataset into a single json file compatible with coco format"""

        for phase, im_paths in self.splits.items():  # Train and Val
            cnt_images = 0
            cnt_instances = 0
            cnt_kps = [0] * 66
            self.initiate_json()  # Initiate json file at each phase

            # save only 50 images
            if self.sample:
                im_paths = im_paths[:50]
            if self.split_images:
                path_dir = (os.path.join(self.dir_out_im, phase))
                os.makedirs(path_dir, exist_ok=True)
                assert not os.listdir(path_dir), "Directory to save images is not empty. " \
                    "Remove flag --split_images ?"
            elif self.single_sample:
                im_paths = self.splits['train'][:1]
                print(f'Single sample for train/val:{im_paths}')

            for im_path in im_paths:
                im_size, im_name, im_id = self._process_image(im_path)
                cnt_images += 1

                # Process its annotations
                txt_paths = glob.glob(os.path.join(self.dir_dataset, 'keypoints', im_name,
                                                   im_name + '*.txt'))
                for txt_path in txt_paths:
                    data = np.loadtxt(txt_path, delimiter='\t', ndmin=2)
                    cnt_kps = self._process_annotation(data, txt_path, im_size, im_id, cnt_kps)
                    cnt_instances += 1

                # Split the image in a new folder
                if self.split_images:
                    dst = os.path.join(self.dir_out_im, phase, os.path.split(im_path)[-1])
                    copyfile(im_path, dst)

                # Add crowd annotations
                mask_path = os.path.join(self.dir_mask, im_name + '.jpg')
                self._process_mask(mask_path, im_id)

                # Count
                if (cnt_images % 1000) == 0:
                    text = ' and copied to new directory' if self.split_images else ''
                    print(f'Parsed {cnt_images} images' + text)

            self.save_json_files(phase)
            print(f'\nPhase:{phase}')
            print(f'Average number of keypoints labelled: {sum(cnt_kps) / cnt_instances:.1f} / 66')
            print(f'JSON files directory:  {self.dir_out_ann}')
            print(f'Saved {cnt_instances} instances over {cnt_images} images ')
            if self.histogram:
                histogram(cnt_kps)

    def save_json_files(self, phase):
        for j_file, n_kps in [(self.json_file_24, 24), (self.json_file_66, 66)]:
            name = 'apollo_keypoints_' + str(n_kps) + '_'
            if self.sample:
                name = name + 'sample_'
            elif self.single_sample:
                name = name + 'single_sample_'

            path_json = os.path.join(self.dir_out_ann, name + phase + '.json')
            with open(path_json, 'w') as outfile:
                json.dump(j_file, outfile)

    def _process_image(self, im_path):
        """Update image field in json file"""
        file_name = os.path.basename(im_path)
        im_name = os.path.splitext(file_name)[0]
        im_id = int(im_name.split(sep='_')[1])  # Numeric code in the image
        im = Image.open(im_path)
        width, height = im.size
        dict_ann = {
            'coco_url': "unknown",
            'file_name': file_name,
            'id': im_id,
            'license': 1,
            'date_captured': "unknown",
            'width': width,
            'height': height}
        self.json_file_24["images"].append(dict_ann)
        self.json_file_66["images"].append(dict_ann)
        return (width, height), im_name, im_id

    def _process_mask(self, mask_path, im_id):
        """Mask crowd annotations"""
        if cv2 is None:
            raise Exception('OpenCV')

        assert os.path.isfile(mask_path), mask_path
        im_gray = cv2.imread(mask_path, cv2.IMREAD_GRAYSCALE)
        blur = cv2.GaussianBlur(im_gray, (0, 0), sigmaX=3, sigmaY=3, borderType=cv2.BORDER_DEFAULT)
        contours, _ = cv2.findContours(blur, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

        for idx, mask in enumerate(contours):
            box = cv2.boundingRect(mask)
            mask_id = int(str(im_id) + '00' + str(idx))  # different id from crowds
            dict_mask = {
                'image_id': im_id,
                'category_id': 1,
                'iscrowd': 1,
                'id': mask_id,
                'area': box[2] * box[3],
                'bbox': box,
                'num_keypoints': 0,
                'keypoints': [],
                'segmentation': []}
            self.json_file_24["annotations"].append(dict_mask)
            self.json_file_66["annotations"].append(dict_mask)

    def _process_annotation(self, all_kps, txt_path, im_size, im_id, cnt_kps):
        """Process single instance"""

        # Enlarge box
        box_tight = [np.min(all_kps[:, 1]), np.min(all_kps[:, 2]),
                     np.max(all_kps[:, 1]), np.max(all_kps[:, 2])]
        w, h = box_tight[2] - box_tight[0], box_tight[3] - box_tight[1]
        x_o = max(box_tight[0] - 0.1 * w, 0)
        y_o = max(box_tight[1] - 0.1 * h, 0)
        x_i = min(box_tight[0] + 1.1 * w, im_size[0])
        y_i = min(box_tight[1] + 1.1 * h, im_size[1])
        box = [int(x_o), int(y_o), int(x_i - x_o), int(y_i - y_o)]  # (x, y, w, h)

        txt_id = os.path.splitext(txt_path.split(sep='_')[-1])[0]
        car_id = int(str(im_id) + str(int(txt_id)))  # include the specific annotation id
        kps, num = self._transform_keypoints_24(all_kps)
        self.json_file_24["annotations"].append({
            'image_id': im_id,
            'category_id': 1,
            'iscrowd': 0,
            'id': car_id,
            'area': box[2] * box[3],
            'bbox': box,
            'num_keypoints': num,
            'keypoints': kps,
            'segmentation': []})

        kps, num = self._transform_keypoints_66(all_kps)
        self.json_file_66["annotations"].append({
            'image_id': im_id,
            'category_id': 1,
            'iscrowd': 0,
            'id': car_id,
            'area': box[2] * box[3],
            'bbox': box,
            'num_keypoints': num,
            'keypoints': kps,
            'segmentation': []})
        # Stats
        for num in all_kps[:, 0]:
            cnt_kps[int(num)] += 1
        return cnt_kps

    def _transform_keypoints_24(self, kps):
        """
        24 keypoint version
        Map, filter keypoints and add visibility
        :array of [[#, x, y], ...]
        :return  [x, y, visibility, x, y, visibility, .. ]
        """
        kps_out = np.zeros((len(CAR_KEYPOINTS_24), 3))
        cnt = 0
        for kp in kps:
            n = self.map_sk[int(kp[0])]
            if n < 100:  # Filter extra keypoints
                kps_out[n, 0] = kp[1]
                kps_out[n, 1] = kp[2]
                kps_out[n, 2] = 2
                cnt += 1
        kps_out = list(kps_out.reshape((-1,)))
        return kps_out, cnt

    @staticmethod
    def _transform_keypoints_66(kps):
        """
        66 keypoint version
        Add visibility
        :array of [[#, x, y], ...]
        :return  [x, y, visibility, x, y, visibility, .. ]
        """
        kps_out = np.zeros((len(CAR_KEYPOINTS_66), 3))
        cnt = 0
        for kp in kps:
            n = int(kp[0])
            kps_out[n, 0] = kp[1]
            kps_out[n, 1] = kp[2]
            kps_out[n, 2] = 2
            cnt += 1
        kps_out = list(kps_out.reshape((-1,)))
        return kps_out, cnt

    def initiate_json(self):
        """
        Initiate Json for training and val phase for the 24 kp and the 66 kp version
        """
        for j_file, n_kp in [(self.json_file_24, 24), (self.json_file_66, 66)]:
            j_file["info"] = dict(url="https://github.com/openpifpaf/openpifpaf",
                                  date_created=time.strftime("%a, %d %b %Y %H:%M:%S +0000",
                                                             time.localtime()),
                                  description=("Conversion of ApolloCar3D dataset into MS-COCO"
                                               " format with {n_kp} keypoints"))

            skel = CAR_SKELETON_24 if n_kp == 24 else CAR_SKELETON_66
            car_kps = CAR_KEYPOINTS_24 if n_kp == 24 else CAR_KEYPOINTS_66
            j_file["categories"] = [dict(name='car',
                                         id=1,
                                         skeleton=skel,
                                         supercategory='car',
                                         keypoints=car_kps)]
            j_file["images"] = []
            j_file["annotations"] = []


def histogram(cnt_kps):
    bins = np.arange(len(cnt_kps))
    data = np.array(cnt_kps)
    plt.figure()
    plt.title("Distribution of the keypoints")
    plt.bar(bins, data)
    plt.xticks(np.arange(len(cnt_kps), step=5))
    plt.show()


def main():
    args = cli()

    # configure
    ApolloToCoco.sample = args.sample
    ApolloToCoco.single_sample = args.single_sample
    ApolloToCoco.split_images = args.split_images
    ApolloToCoco.histogram = args.histogram

    apollo_coco = ApolloToCoco(args.dir_data, args.dir_out)
    apollo_coco.process()


if __name__ == "__main__":
    main()



import logging

import numpy as np

from openpifpaf.metric.base import Base
from openpifpaf.annotation import Annotation

try:
    import scipy
except ImportError:
    scipy = None

LOG = logging.getLogger(__name__)


class MeanPixelError(Base):
    """
    Calculate mean pixel error and detection rate for a given image
    and category in an "all-vs-all setting"
    """
    predictions = []
    image_ids = []
    errors = []  # mean pixel errors
    detections = []  # detection rate
    errors_scaled = []  # mean pixel errors
    detections_scaled = []  # detection rate
    px_ref = 368  # CPM crop size in pixels

    def accumulate(self, predictions, image_meta, *, ground_truth=None):
        errors = []
        detections = []
        errors_scaled = []
        detections_scaled = []

        # Filter ground-truth
        for annotation in ground_truth:
            if not isinstance(annotation, Annotation):
                continue
            indices_gt = np.nonzero(annotation.data[:, 2] > 1.0)
            if indices_gt[0].size <= 3:
                continue
            gts = annotation.data[indices_gt, 0:2].squeeze()
            width = float(annotation.fixed_bbox[2])
            height = float(annotation.fixed_bbox[3])
            scale = np.array([self.px_ref / width, self.px_ref / height]).reshape(1, 2)

            # Evaluate each keypoint
            for idx, gt in zip(indices_gt[0], gts):
                preds = np.array([p.data[idx] for p in predictions]).reshape(-1, 3)[:, 0:2]
                if preds.size <= 0:
                    continue
                i = np.argmin(np.linalg.norm(preds - gt, axis=1))
                dist = preds[i:i + 1] - gt
                dist_scaled = dist * scale
                d = float(np.linalg.norm(dist, axis=1))
                d_scaled = float(np.linalg.norm(dist_scaled, axis=1))

                # Prediction correct if error less than 10 pixels
                if d < 10:
                    errors.append(d)
                    detections.append(1)
                else:
                    detections.append(0)
                if d_scaled < 10:
                    errors_scaled.append(d)
                    detections_scaled.append(1)
                else:
                    detections_scaled.append(0)

        # Stats for a single image
        mpe = average(errors)
        mpe_scaled = average(errors_scaled)
        det_rate = 100 * average(detections)
        det_rate_scaled = 100 * average(detections_scaled)
        LOG.info('Mean Pixel Error (scaled): %s (%s)    Det. Rate (scaled): %s (%s)',
                 str(mpe)[:4], str(mpe_scaled)[:4], str(det_rate)[:4], str(det_rate_scaled)[:4])

        # Accumulate stats
        self.errors.extend(errors)
        self.detections.extend(detections)
        self.errors_scaled.extend(errors_scaled)
        self.detections_scaled.extend(detections_scaled)

    def write_predictions(self, filename, *, additional_data=None):
        raise NotImplementedError

    def stats(self):
        mpe = average(self.errors)
        mpe_scaled = average(self.errors_scaled)
        det_rate = 100 * average(self.detections)
        det_rate_scaled = 100 * average(self.detections_scaled)
        LOG.info('Final Results: \nMean Pixel Error [scaled] : %f [%f] '
                 '\nDetection Rate [scaled]: %f [%f]',
                 mpe, mpe_scaled, det_rate, det_rate_scaled)
        data = {
            'stats': [mpe, mpe_scaled, det_rate, det_rate_scaled],
            'text_labels': ['Mean Pixel Error',
                            'Mean Pixel Error Scaled',
                            'Detection Rate [%]',
                            'Detection Rate Scaled[%]'],
        }
        return data


def hungarian_matching(gts, predictions, thresh=0.5):
    cost = np.zeros((len(gts), len(predictions)))

    for i, (dg, vg) in enumerate(gts):
        for j, pred in enumerate(predictions):
            p = np.array(pred.data)
            dp = p[:, 0:2][vg > 1.0]
            vp = p[:, 2][vg > 1.0]

            dp[vp < thresh] = -100
            dp[vp < thresh] = -100

            # measure the per-keypoint distance
            distances = np.clip(np.linalg.norm(dp - dg, axis=1), 0, 10)
            cost[i, j] = float(np.mean(distances))

    assert np.max(cost) < 11
    row, cols = scipy.optimize.linear_sum_assignment(cost)
    return row, cols, cost


def average(my_list, *, empty_value=0.0):
    """calculate mean of a list"""
    if not my_list:
        return empty_value

    return sum(my_list) / float(len(my_list))



import logging
import numpy as np

LOG = logging.getLogger(__name__)


class Crafted:
    """Compute hand crafted distance between a track and a new pose candidate.

    Ingredients:
    * compare to last pose in track and previous poses in case of temp corruption
    * require low distance for points that have high confidence in both poses (>=3 points)
    * "high confidence" is a dynamic measure dependent on the past track
    * penalize crappy tracks
    * penalize crappy poses
    """

    invisible_penalty = 110.0

    def __init__(self):
        self.valid_keypoints = None

    def __call__(self, frame_number, pose, track, track_is_good):
        return min((
            self.distance(frame_number, pose, track, track_is_good),
            self.distance(frame_number, pose, track, track_is_good, -4),
            self.distance(frame_number, pose, track, track_is_good, -8),
            self.distance(frame_number, pose, track, track_is_good, -12),
        ))

    # pylint: disable=too-many-return-statements,too-many-branches
    def distance(self, frame_number, pose, track, track_is_good, track_frame=-1):
        last_track_frame = track.frame_pose[-1][0]
        skipped_frames = frame_number - last_track_frame - 1
        assert skipped_frames >= 0
        if skipped_frames > 12:
            return 1000.0

        # correct track_frame with skipped_frames
        track_frame += skipped_frames
        if track_frame > -1:
            return 1000.0
        if len(track.frame_pose) < -track_frame:
            return 1000.0

        pose1 = pose.data[self.valid_keypoints]
        pose2 = track.frame_pose[track_frame][1].data[self.valid_keypoints]

        keypoint_scores = pose1[:, 2] * pose2[:, 2]
        kps_order = np.argsort(keypoint_scores)[::-1]
        if pose1[kps_order[2], 2] < 0.05 or pose2[kps_order[2], 2] < 0.05:
            return 1000.0
        pose1_center = np.mean(pose1[kps_order[:3], :2], axis=0)
        pose1_centered = np.copy(pose1)
        pose1_centered[:, :2] -= pose1_center
        pose2_center = np.mean(pose2[kps_order[:3], :2], axis=0)
        pose2_centered = np.copy(pose2)
        pose2_centered[:, :2] -= pose2_center
        center_distance = np.linalg.norm(pose2_center - pose1_center)

        kps_distances = np.linalg.norm(pose2_centered[:, :2] - pose1_centered[:, :2], axis=1)
        kps_distances = np.clip(kps_distances, 0.0, self.invisible_penalty)
        kps_distances[pose1[:, 2] < 0.05] = self.invisible_penalty
        kps_distances[pose2[:, 2] < 0.05] = self.invisible_penalty
        kps_distance_centered = np.mean(kps_distances)

        crappy_track_penalty = 0.0
        if len(track.frame_pose) < 4:
            crappy_track_penalty = 40.0
        elif len(track.frame_pose) < 8:
            crappy_track_penalty = 8.0
        if not track_is_good:
            crappy_track_penalty = max(crappy_track_penalty, 8.0)

        crappy_pose_penalty = 0.0
        if pose.score() < 0.2:
            crappy_pose_penalty = 40.0
        elif pose.score() < 0.5:
            crappy_pose_penalty = 8.0

        # skipping frames cost
        skipped_frame_cost = 40.0 if track_frame < -1 else 0.0

        return (
            center_distance / 10.0
            + kps_distance_centered
            + crappy_track_penalty
            + crappy_pose_penalty
            + skipped_frame_cost
        )



import copy
import logging

from .base import Base
from ..annotation import Annotation, AnnotationCrowd
from .. import headmeta, show

try:
    import matplotlib.cm
    CMAP_ORANGES_NAN = copy.copy(matplotlib.cm.get_cmap('Oranges'))
    CMAP_ORANGES_NAN.set_bad('white', alpha=0.5)
except ImportError:
    CMAP_ORANGES_NAN = None

LOG = logging.getLogger(__name__)


class Cif(Base):
    """Visualize a CIF field."""

    def __init__(self, meta: headmeta.Cif):
        super().__init__(meta.name)
        self.meta = meta
        keypoint_painter = show.KeypointPainter(monocolor_connections=True)
        self.annotation_painter = show.AnnotationPainter(painters={'Annotation': keypoint_painter})

    def targets(self, field, *, annotation_dicts):
        assert self.meta.keypoints is not None
        assert self.meta.draw_skeleton is not None

        annotations = [
            Annotation(
                keypoints=self.meta.keypoints,
                skeleton=self.meta.draw_skeleton,
                sigmas=self.meta.sigmas,
                score_weights=self.meta.score_weights
            ).set(
                ann['keypoints'], fixed_score='', fixed_bbox=ann['bbox'])
            if not ann['iscrowd']
            else AnnotationCrowd(['keypoints']).set(1, ann['bbox'])
            for ann in annotation_dicts
        ]

        self._confidences(field[:, 0])
        self._regressions(field[:, 1:3], field[:, 4], annotations=annotations)

    def predicted(self, field):
        self._confidences(field[:, 1])
        self._regressions(field[:, 2:4], field[:, 4],
                          annotations=self._ground_truth,
                          confidence_fields=field[:, 1],
                          uv_is_offset=False)

    def _confidences(self, confidences):
        for f in self.indices('confidence'):
            LOG.debug('%s', self.meta.keypoints[f])

            with self.image_canvas(self.processed_image(), margin=[0.0, 0.01, 0.05, 0.01]) as ax:
                im = ax.imshow(self.scale_scalar(confidences[f], self.meta.stride),
                               alpha=0.9, vmin=0.0, vmax=1.0, cmap=CMAP_ORANGES_NAN)
                self.colorbar(ax, im)

    def _regressions(self, regression_fields, scale_fields, *,
                     annotations=None, confidence_fields=None, uv_is_offset=True):
        for f in self.indices('regression'):
            LOG.debug('%s', self.meta.keypoints[f])
            confidence_field = confidence_fields[f] if confidence_fields is not None else None

            with self.image_canvas(self.processed_image(), margin=[0.0, 0.01, 0.05, 0.01]) as ax:
                show.white_screen(ax, alpha=0.5)
                if annotations:
                    self.annotation_painter.annotations(ax, annotations, color='lightgray')
                q = show.quiver(ax,
                                regression_fields[f, :2],
                                confidence_field=confidence_field,
                                xy_scale=self.meta.stride, uv_is_offset=uv_is_offset,
                                cmap='Oranges', clim=(0.5, 1.0), width=0.001)
                show.boxes(ax, scale_fields[f] / 2.0,
                           confidence_field=confidence_field,
                           regression_field=regression_fields[f, :2],
                           xy_scale=self.meta.stride, cmap='Oranges', fill=False,
                           regression_field_is_offset=uv_is_offset)
                if f in self.indices('margin', with_all=False):
                    show.margins(ax, regression_fields[f, :6], xy_scale=self.meta.stride)

                self.colorbar(ax, q)



"""Context for doctests."""

import numpy
import pytest

import openpifpaf


@pytest.fixture(autouse=True)
def add_np(doctest_namespace):
    doctest_namespace['np'] = numpy


@pytest.fixture(autouse=True)
def add_openpifpaf(doctest_namespace):
    doctest_namespace['openpifpaf'] = openpifpaf



import copy
import logging

import PIL

from ..preprocess import Preprocess

LOG = logging.getLogger(__name__)


class SamplePairing(Preprocess):
    """Sample Pairing

    Implements:
    @article{inoue2018data,
        title={Data augmentation by pairing samples for images classification},
        author={Inoue, Hiroshi},
        journal={arXiv preprint arXiv:1801.02929},
        year={2018}
    }
    This was originally published for classification and adapted here for
    pose estimation.
    """

    def __init__(self):
        self.previous_images = None
        self.previous_all_annotations = []

    def __call__(self, original_images, original_all_anns, metas):
        images = original_images
        all_anns = copy.deepcopy(original_all_anns)

        if self.previous_images is not None:
            # image
            images = [
                PIL.Image.blend(current_image, previous_image, 0.5)
                for current_image, previous_image in zip(images, self.previous_images)
            ]

            # annotations
            for current_anns, previous_anns in zip(all_anns, self.previous_all_annotations):
                current_anns += previous_anns

            # meta untouched

        self.previous_images = original_images
        self.previous_all_annotations = original_all_anns
        return images, all_anns, metas



"""Head networks."""

import argparse
import functools
import logging
import math

import torch

from .. import headmeta

LOG = logging.getLogger(__name__)


@functools.lru_cache(maxsize=16)
def index_field_torch(shape, *, device=None, unsqueeze=(0, 0)):
    assert len(shape) == 2
    xy = torch.empty((2, shape[0], shape[1]), device=device)
    xy[0] = torch.arange(shape[1], device=device)
    xy[1] = torch.arange(shape[0], device=device).unsqueeze(1)

    for dim in unsqueeze:
        xy = torch.unsqueeze(xy, dim)

    return xy


class PifHFlip(torch.nn.Module):
    def __init__(self, keypoints, hflip):
        super().__init__()

        flip_indices = torch.LongTensor([
            keypoints.index(hflip[kp_name]) if kp_name in hflip else kp_i
            for kp_i, kp_name in enumerate(keypoints)
        ])
        LOG.debug('hflip indices: %s', flip_indices)
        self.register_buffer('flip_indices', flip_indices)

    def forward(self, *args):
        out = []
        for field in args:
            field = torch.index_select(field, 1, self.flip_indices)
            field = torch.flip(field, dims=[len(field.shape) - 1])
            out.append(field)

        # flip the x-coordinate of the vector component
        out[1][:, :, 0, :, :] *= -1.0

        return out


class PafHFlip(torch.nn.Module):
    def __init__(self, keypoints, skeleton, hflip):
        super().__init__()
        skeleton_names = [
            (keypoints[j1 - 1], keypoints[j2 - 1])
            for j1, j2 in skeleton
        ]
        flipped_skeleton_names = [
            (hflip[j1] if j1 in hflip else j1, hflip[j2] if j2 in hflip else j2)
            for j1, j2 in skeleton_names
        ]
        LOG.debug('skeleton = %s, flipped_skeleton = %s',
                  skeleton_names, flipped_skeleton_names)

        flip_indices = list(range(len(skeleton)))
        reverse_direction = []
        for paf_i, (n1, n2) in enumerate(skeleton_names):
            if (n1, n2) in flipped_skeleton_names:
                flip_indices[paf_i] = flipped_skeleton_names.index((n1, n2))
            if (n2, n1) in flipped_skeleton_names:
                flip_indices[paf_i] = flipped_skeleton_names.index((n2, n1))
                reverse_direction.append(paf_i)
        LOG.debug('hflip indices: %s, reverse: %s', flip_indices, reverse_direction)

        self.register_buffer('flip_indices', torch.LongTensor(flip_indices))
        self.register_buffer('reverse_direction', torch.LongTensor(reverse_direction))

    def forward(self, *args):
        out = []
        for field in args:
            field = torch.index_select(field, 1, self.flip_indices)
            field = torch.flip(field, dims=[len(field.shape) - 1])
            out.append(field)

        # flip the x-coordinate of the vector components
        out[1][:, :, 0, :, :] *= -1.0
        out[2][:, :, 0, :, :] *= -1.0

        # reverse direction
        for paf_i in self.reverse_direction:
            cc = torch.clone(out[1][:, paf_i])
            out[1][:, paf_i] = out[2][:, paf_i]
            out[2][:, paf_i] = cc

        return out


class HeadNetwork(torch.nn.Module):
    """Base class for head networks.

    :param meta: head meta instance to configure this head network
    :param in_features: number of input features which should be equal to the
        base network's output features
    """
    def __init__(self, meta: headmeta.Base, in_features: int):
        super().__init__()
        self.meta = meta
        self.in_features = in_features

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        """Command line interface (CLI) to extend argument parser."""

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""

    def forward(self, x):
        raise NotImplementedError


class CompositeField3(HeadNetwork):
    dropout_p = 0.0
    inplace_ops = True

    def __init__(self,
                 meta: headmeta.Base,
                 in_features, *,
                 kernel_size=1, padding=0, dilation=1):
        super().__init__(meta, in_features)

        LOG.debug('%s config: fields = %d, confidences = %d, vectors = %d, scales = %d '
                  'kernel = %d, padding = %d, dilation = %d',
                  meta.name, meta.n_fields, meta.n_confidences, meta.n_vectors, meta.n_scales,
                  kernel_size, padding, dilation)

        self.dropout = torch.nn.Dropout2d(p=self.dropout_p)

        # convolution
        out_features = meta.n_fields * (meta.n_confidences + meta.n_vectors * 3 + meta.n_scales)
        self.conv = torch.nn.Conv2d(in_features, out_features * (meta.upsample_stride ** 2),
                                    kernel_size, padding=padding, dilation=dilation)

        # upsample
        assert meta.upsample_stride >= 1
        self.upsample_op = None
        if meta.upsample_stride > 1:
            self.upsample_op = torch.nn.PixelShuffle(meta.upsample_stride)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('CompositeField3')
        group.add_argument('--cf3-dropout', default=cls.dropout_p, type=float,
                           help='[experimental] zeroing probability of feature in head input')
        assert cls.inplace_ops
        group.add_argument('--cf3-no-inplace-ops', dest='cf3_inplace_ops',
                           default=True, action='store_false',
                           help='alternative graph without inplace ops')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.dropout_p = args.cf3_dropout
        cls.inplace_ops = args.cf3_inplace_ops

    @property
    def sparse_task_parameters(self):
        return [self.conv.weight]

    def forward(self, x):  # pylint: disable=arguments-differ
        x = self.dropout(x)
        x = self.conv(x)
        # upscale
        if self.upsample_op is not None:
            x = self.upsample_op(x)
            low_cut = (self.meta.upsample_stride - 1) // 2
            high_cut = math.ceil((self.meta.upsample_stride - 1) / 2.0)
            if self.training:
                # negative axes not supported by ONNX TensorRT
                x = x[:, :, low_cut:-high_cut, low_cut:-high_cut]
            else:
                # the int() forces the tracer to use static shape
                x = x[:, :, low_cut:int(x.shape[2]) - high_cut, low_cut:int(x.shape[3]) - high_cut]

        # Extract some shape parameters once.
        # Convert to int so that shape is constant in ONNX export.
        x_size = x.size()
        batch_size = x_size[0]
        feature_height = int(x_size[2])
        feature_width = int(x_size[3])

        x = x.view(
            batch_size,
            self.meta.n_fields,
            self.meta.n_confidences + self.meta.n_vectors * 3 + self.meta.n_scales,
            feature_height,
            feature_width
        )

        if not self.training and self.inplace_ops:
            # classification
            classes_x = x[:, :, 0:self.meta.n_confidences]
            torch.sigmoid_(classes_x)

            # regressions x: add index
            if self.meta.n_vectors > 0:
                index_field = index_field_torch((feature_height, feature_width), device=x.device)
                first_reg_feature = self.meta.n_confidences
                for i, do_offset in enumerate(self.meta.vector_offsets):
                    if not do_offset:
                        continue
                    reg_x = x[:, :, first_reg_feature + i * 2:first_reg_feature + (i + 1) * 2]
                    reg_x.add_(index_field)

            # scale
            first_scale_feature = self.meta.n_confidences + self.meta.n_vectors * 3
            scales_x = x[:, :, first_scale_feature:first_scale_feature + self.meta.n_scales]
            scales_x[:] = torch.nn.functional.softplus(scales_x)

            # remove width in the middle and add one to the front (v4 style)
            first_width_feature = self.meta.n_confidences + self.meta.n_vectors * 2
            x = torch.cat([
                x[:, :, first_width_feature:first_width_feature + 1],
                x[:, :, :first_width_feature],
                x[:, :, self.meta.n_confidences + self.meta.n_vectors * 3:],
            ], dim=2)
        elif not self.training and not self.inplace_ops:
            # TODO: CoreMLv4 does not like strided slices.
            # Strides are avoided when switching the first and second dim
            # temporarily.
            x = torch.transpose(x, 1, 2)

            # classification
            classes_x = x[:, 0:self.meta.n_confidences]
            classes_x = torch.sigmoid(classes_x)

            # regressions x
            first_reg_feature = self.meta.n_confidences
            regs_x = [
                x[:, first_reg_feature + i * 2:first_reg_feature + (i + 1) * 2]
                for i in range(self.meta.n_vectors)
            ]
            # regressions x: add index
            index_field = index_field_torch(
                (feature_height, feature_width), device=x.device, unsqueeze=(1, 0))
            # TODO: coreml export does not work with the index_field creation in the graph.
            index_field = torch.from_numpy(index_field.numpy())
            regs_x = [reg_x + index_field if do_offset else reg_x
                      for reg_x, do_offset in zip(regs_x, self.meta.vector_offsets)]

            # regressions logb
            first_reglogb_feature = self.meta.n_confidences + self.meta.n_vectors * 2
            single_reg_logb = x[:, first_reglogb_feature:first_reglogb_feature + 1]

            # scale
            first_scale_feature = self.meta.n_confidences + self.meta.n_vectors * 3
            scales_x = x[:, first_scale_feature:first_scale_feature + self.meta.n_scales]
            scales_x = torch.nn.functional.softplus(scales_x)

            # concat with width in front (v4 style)
            x = torch.cat([single_reg_logb, classes_x, *regs_x, scales_x], dim=1)

            # TODO: CoreMLv4 problem (see above).
            x = torch.transpose(x, 1, 2)

        return x


class CompositeField4(HeadNetwork):
    dropout_p = 0.0
    inplace_ops = True

    def __init__(self,
                 meta: headmeta.Base,
                 in_features, *,
                 kernel_size=1, padding=0, dilation=1):
        super().__init__(meta, in_features)

        LOG.debug('%s config: fields = %d, confidences = %d, vectors = %d, scales = %d '
                  'kernel = %d, padding = %d, dilation = %d',
                  meta.name, meta.n_fields, meta.n_confidences, meta.n_vectors, meta.n_scales,
                  kernel_size, padding, dilation)

        self.dropout = torch.nn.Dropout2d(p=self.dropout_p)

        # convolution
        self.n_components = 1 + meta.n_confidences + meta.n_vectors * 2 + meta.n_scales
        self.conv = torch.nn.Conv2d(
            in_features, meta.n_fields * self.n_components * (meta.upsample_stride ** 2),
            kernel_size, padding=padding, dilation=dilation,
        )

        # upsample
        assert meta.upsample_stride >= 1
        self.upsample_op = None
        if meta.upsample_stride > 1:
            self.upsample_op = torch.nn.PixelShuffle(meta.upsample_stride)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('CompositeField4')
        group.add_argument('--cf4-dropout', default=cls.dropout_p, type=float,
                           help='[experimental] zeroing probability of feature in head input')
        assert cls.inplace_ops
        group.add_argument('--cf4-no-inplace-ops', dest='cf4_inplace_ops',
                           default=True, action='store_false',
                           help='alternative graph without inplace ops')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.dropout_p = args.cf4_dropout
        cls.inplace_ops = args.cf4_inplace_ops

    @property
    def sparse_task_parameters(self):
        return [self.conv.weight]

    def forward(self, x):  # pylint: disable=arguments-differ
        x = self.dropout(x)
        x = self.conv(x)
        # upscale
        if self.upsample_op is not None:
            x = self.upsample_op(x)
            low_cut = (self.meta.upsample_stride - 1) // 2
            high_cut = math.ceil((self.meta.upsample_stride - 1) / 2.0)
            if self.training:
                # negative axes not supported by ONNX TensorRT
                x = x[:, :, low_cut:-high_cut, low_cut:-high_cut]
            else:
                # the int() forces the tracer to use static shape
                x = x[:, :, low_cut:int(x.shape[2]) - high_cut, low_cut:int(x.shape[3]) - high_cut]

        # Extract some shape parameters once.
        # Convert to int so that shape is constant in ONNX export.
        x_size = x.size()
        batch_size = x_size[0]
        feature_height = int(x_size[2])
        feature_width = int(x_size[3])

        x = x.view(
            batch_size,
            self.meta.n_fields,
            self.n_components,
            feature_height,
            feature_width
        )

        if not self.training and self.inplace_ops:
            # classification
            classes_x = x[:, :, 1:1 + self.meta.n_confidences]
            torch.sigmoid_(classes_x)

            # regressions x: add index
            if self.meta.n_vectors > 0:
                index_field = index_field_torch((feature_height, feature_width), device=x.device)
                first_reg_feature = 1 + self.meta.n_confidences
                for i, do_offset in enumerate(self.meta.vector_offsets):
                    if not do_offset:
                        continue
                    reg_x = x[:, :, first_reg_feature + i * 2:first_reg_feature + (i + 1) * 2]
                    reg_x.add_(index_field)

            # scale
            first_scale_feature = 1 + self.meta.n_confidences + self.meta.n_vectors * 2
            scales_x = x[:, :, first_scale_feature:first_scale_feature + self.meta.n_scales]
            scales_x[:] = torch.nn.functional.softplus(scales_x)
        elif not self.training and not self.inplace_ops:
            # TODO: CoreMLv4 does not like strided slices.
            # Strides are avoided when switching the first and second dim
            # temporarily.
            x = torch.transpose(x, 1, 2)

            # width
            width_x = x[:, 0:1]

            # classification
            classes_x = x[:, 1:1 + self.meta.n_confidences]
            classes_x = torch.sigmoid(classes_x)

            # regressions x
            first_reg_feature = 1 + self.meta.n_confidences
            regs_x = [
                x[:, first_reg_feature + i * 2:first_reg_feature + (i + 1) * 2]
                for i in range(self.meta.n_vectors)
            ]
            # regressions x: add index
            index_field = index_field_torch(
                (feature_height, feature_width), device=x.device, unsqueeze=(1, 0))
            # TODO: coreml export does not work with the index_field creation in the graph.
            index_field = torch.from_numpy(index_field.numpy())
            regs_x = [reg_x + index_field if do_offset else reg_x
                      for reg_x, do_offset in zip(regs_x, self.meta.vector_offsets)]

            # scale
            first_scale_feature = 1 + self.meta.n_confidences + self.meta.n_vectors * 2
            scales_x = x[:, first_scale_feature:first_scale_feature + self.meta.n_scales]
            scales_x = torch.nn.functional.softplus(scales_x)

            # concat
            x = torch.cat([width_x, classes_x, *regs_x, scales_x], dim=1)

            # TODO: CoreMLv4 problem (see above).
            x = torch.transpose(x, 1, 2)

        return x



import numpy as np


class Oks:
    """Compute OKS distance between a track and a new pose candidate.

    From http://cocodataset.org/#keypoints-eval:
    OKS = Σi[exp(-di2/2s2κi2)δ(vi>0)] / Σi[δ(vi>0)]
    with κi=2σi, s is object scale and sigma is the keypoint sigma.

    Ingredients:
    * compare to last pose in track and previous poses in case of temp corruption
    * require low distance for points that have high confidence in both poses (>=3 points)
    * "high confidence" is a dynamic measure dependent on the past track
    * penalize crappy tracks
    * penalize crappy poses
    """
    inflate = 1.0

    def __init__(self, *, track_frames=None):
        if track_frames is None:
            track_frames = [-1]
        assert all(t < 0 for t in track_frames)

        self.track_frames = track_frames
        self.valid_keypoints = None
        self.sigmas = None

    def __call__(self, frame_number, pose, track, track_is_good):
        return min(
            self.distance(frame_number, pose, track, track_is_good, track_frame)
            for track_frame in self.track_frames
        )

    @staticmethod
    def scale(pose):
        pose = pose[pose[:, 2] > 0.0]
        area = (pose[:, 0].max() - pose[:, 0].min()) * (pose[:, 1].max() - pose[:, 1].min())
        return np.sqrt(area)

    def distance(self, frame_number, pose, track, track_is_good, track_frame=-1):  # pylint: disable=unused-argument
        last_track_frame = track.frame_pose[-1][0]
        skipped_frames = frame_number - last_track_frame - 1
        assert skipped_frames >= 0
        if skipped_frames > 12:
            return 1000.0

        # correct track_frame with skipped_frames
        track_frame += skipped_frames
        if track_frame > -1:
            return 1000.0
        if len(track.frame_pose) < -track_frame:
            return 1000.0

        pose1 = pose.data[self.valid_keypoints]
        pose2 = track.frame_pose[track_frame][1].data[self.valid_keypoints]
        visible = np.logical_and(pose1[:, 2] > 0.0, pose2[:, 2] > 0.0)
        if not np.any(visible):
            return 1000.0
        scale = 0.5 * (self.scale(pose1) + self.scale(pose2))
        scale = max(1.0, scale)

        d = np.linalg.norm((pose2[:, :2] - pose1[:, :2]), axis=1)
        k = 2.0 * self.sigmas[self.valid_keypoints] * self.inflate  # pylint: disable=unsubscriptable-object
        g = np.exp(-0.5 * d**2 / (scale**2 * k**2))
        oks = np.mean(g[visible])

        return 110.0 * (1.0 - oks)



import copy
import logging

import numpy as np
import torch

from ..preprocess import Preprocess
from ..crop import Crop as SingleImageCrop

LOG = logging.getLogger(__name__)


class Crop(Preprocess):
    def __init__(self, long_edge, *, use_area_of_interest=True, max_shift=0):
        self.long_edge = long_edge
        self.use_area_of_interest = use_area_of_interest
        self.max_shift = max_shift

    def __call__(self, images, all_anns, metas):
        metas = copy.deepcopy(metas)
        all_anns = copy.deepcopy(all_anns)

        if self.use_area_of_interest:
            # crop around the interesting area in the past frame to
            # train whether the pose continues or not (unless the past
            # frame is blank)
            if all_anns[0] and not all_anns[1]:
                valid_area = metas[0]['valid_area']
                area_of_interest = SingleImageCrop.area_of_interest(
                    all_anns[0], valid_area)
            else:
                valid_area = metas[1]['valid_area']
                area_of_interest = SingleImageCrop.area_of_interest(
                    all_anns[1], valid_area)
        else:
            valid_area = metas[0]['valid_area']
            area_of_interest = valid_area

        new_images = []
        new_anns = []
        new_metas = []

        cam_shift = (torch.rand(2).numpy() - 0.5) * 2.0 * self.max_shift
        LOG.debug('max shift = %s, this shift = %s', self.max_shift, cam_shift)
        for image, anns, meta in zip(images, all_anns, metas):
            LOG.debug('meta group = %d', meta['group_i'])
            original_valid_area = meta['valid_area'].copy()
            with torch.random.fork_rng(devices=[]):
                image, anns, ltrb = self.crop(
                    image, anns, valid_area, area_of_interest, cam_shift * meta.get('group_i', 1.0))
                meta['offset'] += ltrb[:2]

                new_wh = image.size
                LOG.debug('valid area before crop of %s: %s', ltrb, original_valid_area)
                # process crops from left and top
                meta['valid_area'][:2] = np.maximum(0.0, original_valid_area[:2] - ltrb[:2])
                # process crops from right and bottom
                new_rb_corner = original_valid_area[:2] + original_valid_area[2:] - ltrb[:2]
                new_rb_corner = np.maximum(0.0, new_rb_corner)
                new_rb_corner = np.minimum(new_wh, new_rb_corner)
                meta['valid_area'][2:] = new_rb_corner - meta['valid_area'][:2]
                LOG.debug('valid area after crop: %s', meta['valid_area'])

                # clip bounding boxes
                for ann in anns:
                    unclipped_bbox = ann['bbox'].copy()
                    ann['bbox'][:2] = np.maximum(meta['valid_area'][:2], ann['bbox'][:2])
                    new_rb = unclipped_bbox[:2] + unclipped_bbox[2:]
                    new_rb = np.maximum(ann['bbox'][:2], new_rb)
                    new_rb = np.minimum(meta['valid_area'][:2] + meta['valid_area'][2:], new_rb)
                    ann['bbox'][2:] = new_rb - ann['bbox'][:2]
                anns = [ann for ann in anns if ann['bbox'][2] > 0.0 and ann['bbox'][3] > 0.0]

            new_images.append(image)
            new_anns.append(anns)
            new_metas.append(meta)

        return new_images, new_anns, new_metas

    def crop(self, image, anns, valid_area, area_of_interest, cam_shift):
        LOG.debug('cam shift = %s', cam_shift)
        w, h = image.size
        x_offset, y_offset = 0, 0
        if w > self.long_edge:
            x_offset = SingleImageCrop.random_location_1d(
                w - 1,
                valid_area[0], valid_area[2],
                area_of_interest[0], area_of_interest[2],
                self.long_edge,
                shift=cam_shift[0],
                fix_inconsistent=True,
            )
        if h > self.long_edge:
            y_offset = SingleImageCrop.random_location_1d(
                h - 1,
                valid_area[1], valid_area[3],
                area_of_interest[1], area_of_interest[3],
                self.long_edge,
                shift=cam_shift[1],
                fix_inconsistent=True,
            )
        LOG.debug('crop offsets (%d, %d)', x_offset, y_offset)

        # crop image
        new_w = min(self.long_edge, w - x_offset)
        new_h = min(self.long_edge, h - y_offset)
        # ltrb might be confusing name:
        # it's the coordinates of the top-left corner and the coordinates
        # of the bottom right corner
        ltrb = (x_offset, y_offset, x_offset + new_w, y_offset + new_h)
        image = image.crop(ltrb)
        LOG.debug('cropped image: %s', ltrb)

        # crop keypoints
        for ann in anns:
            ann['keypoints'][:, 0] -= x_offset
            ann['keypoints'][:, 1] -= y_offset
            ann['bbox'][0] -= x_offset
            ann['bbox'][1] -= y_offset

        return image, anns, np.array(ltrb)



import logging

from .base import Base

LOG = logging.getLogger(__name__)


class Classification(Base):
    def __init__(self, categories):
        self.categories = ['total'] + list(categories)

        # counters: index 0 is the total
        self.gt_counts = [0 for _ in range(len(categories) + 1)]
        self.correct_counts = [0 for _ in range(len(categories) + 1)]

    def accumulate(self, predictions, image_meta, *, ground_truth=None):
        # get categories
        gt_category = ground_truth[0].category_id
        if predictions:
            max_prediction = max(predictions, key=lambda p: p.score)
            pred_category = max_prediction.category_id
        else:
            pred_category = None
        LOG.debug('ground truth = %s, prediction = %s', gt_category, pred_category)

        # add to counts
        self.gt_counts[0] += 1
        self.gt_counts[gt_category] += 1
        if gt_category == pred_category:
            self.correct_counts[0] += 1
            self.correct_counts[gt_category] += 1

    def stats(self):
        return {
            'text_labels': self.categories,
            'stats': [correct / gt_count if gt_count else 0.0
                      for correct, gt_count in zip(self.correct_counts, self.gt_counts)]
        }

    def write_predictions(self, filename, *, additional_data=None):
        raise NotImplementedError



from ..signal import Signal


class LoaderWithReset:
    """Wraps another data loader and emits a signal when a meta key changes.

    This is used for video datasets where one of the keys is the
    video sequence id. When that id changes, this loader emits a 'eval_reset'
    signal that the decoder and a RunningCache can subscribe to to reset
    their internal state.
    """

    def __init__(self, parent, key_to_monitor):
        self.parent = parent
        self.key_to_monitor = key_to_monitor

        self.previous_value = None

    def __iter__(self):
        for images, anns, metas in self.parent:
            value = metas[0][self.key_to_monitor]
            if len(metas) >= 2:
                assert all(m[self.key_to_monitor] == value for m in metas[1:])

            if value != self.previous_value:
                Signal.emit('eval_reset')
                self.previous_value = value

            yield images, anns, metas

    def __len__(self):
        return len(self.parent)



from collections import defaultdict
import datetime
import glob
import json
import logging
import os
import subprocess

import numpy as np
import openpifpaf

from .constants import POSETRACK_INDEX2017TO2018

LOG = logging.getLogger(__name__)


class Posetrack(openpifpaf.metric.Base):
    def __init__(self, *, images, categories,
                 ground_truth=None, output_format='2018'):
        super().__init__()

        self.images_by_file = images
        self.categories_by_file = categories
        self.ground_truth_directory = ground_truth
        self.output_format = output_format

        self.predictions_by_file = defaultdict(list)
        self.output_dir_suffix = '{}-{}'.format(
            output_format,
            datetime.datetime.now().strftime('%y%m%d-%H%M%S'),
        )

        self._written_mot_stats_file = None
        self._written_ap_stats_file = None

    def stats(self):
        if self._written_ap_stats_file is None \
           or self._written_mot_stats_file is None:
            return {
                'stats': [],
                'text_labels': [],
            }

        with open(self._written_mot_stats_file, 'r') as f_mot:
            mot_stats = json.load(f_mot)
        with open(self._written_ap_stats_file, 'r') as f_ap:
            ap_stats = json.load(f_ap)

        mot_index_by_name = {n: int(i) for i, n in mot_stats['names'].items()}
        ap_index_by_name = {n: int(i) for i, n in ap_stats['names'].items()}

        return {
            'stats': [
                mot_stats['mota'][mot_index_by_name['total']],
                0.5 * (ap_stats['ap'][ap_index_by_name['right_wrist']]
                       + ap_stats['ap'][ap_index_by_name['left_wrist']]),
                0.5 * (ap_stats['ap'][ap_index_by_name['right_ankle']]
                       + ap_stats['ap'][ap_index_by_name['left_ankle']]),
                ap_stats['ap'][ap_index_by_name['total']],
            ],
            'text_labels': [
                'MOTA', 'AP_wrists', 'AP_ankles', 'AP',
            ],
        }

    def accumulate(self, predictions, image_meta, *, ground_truth=None):  # pylint: disable=unused-argument
        annotation_file = image_meta['annotation_file']

        # make sure an entry for this annotation file is created
        # even when a file does not have any predictions
        if annotation_file not in self.predictions_by_file:
            self.predictions_by_file[annotation_file] = []

        for ann in predictions:
            keypoints = np.copy(ann.data)

            # remove points outside image
            w, h = image_meta['width_height']
            keypoints[keypoints[:, 0] < 0.0, 2] = 0.0
            keypoints[keypoints[:, 1] < 0.0, 2] = 0.0
            keypoints[keypoints[:, 0] > w - 1, 2] = 0.0
            keypoints[keypoints[:, 1] > h - 1, 2] = 0.0

            # cleanup
            keypoints[:, 2] = np.clip(keypoints[:, 2], 0.0, 1.0)
            keypoints[keypoints[:, 2] == 0.0, :2] = 0.0

            bbox = [float(v) for v in ann.bbox()]
            bbox = [bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]]
            bbox = [round(v, 2) for v in bbox]

            self.predictions_by_file[annotation_file].append({
                'bbox': bbox,
                'keypoints': [round(k, 2) for k in keypoints.reshape(-1).tolist()],
                'track_id': ann.id_,
                'image_id': image_meta['image_id'],
                'image_file': image_meta['file_name'],
                'category_id': 1,
                'scores': [round(s, 2) for s in keypoints[:, 2].tolist()],
                'score': max(0.01, round(ann.score, 2)),
            })

    def _write2018(self, output_dir, annotation_file, *, additional_data=None):
        sequence_name = os.path.basename(annotation_file)
        out_name = '{}/{}'.format(output_dir, sequence_name)
        os.makedirs(os.path.dirname(out_name), exist_ok=True)
        LOG.info('writing %s', out_name)

        data = {
            'images': self.images_by_file[annotation_file],
            'annotations': self.predictions_by_file[annotation_file],
            'categories': self.categories_by_file[annotation_file],
        }
        if additional_data:
            data = dict(**data, **additional_data)
        with open(out_name, 'w') as f:
            json.dump(data, f)
        LOG.info('wrote %s', out_name)

    def _write2017(self, output_dir, annotation_file, *, additional_data=None):
        sequence_name = os.path.basename(annotation_file)
        out_name = '{}/{}'.format(output_dir, sequence_name)
        os.makedirs(os.path.dirname(out_name), exist_ok=True)
        LOG.info('writing %s', out_name)

        data = {
            # 'args': sys.argv,
            'annolist': self._format_2017(
                self.images_by_file[annotation_file],
                self.predictions_by_file[annotation_file],
            ),
            # 'categories': categories[self.annotation_file],
            # 'total_time': round(total_time, 3),
            # 'nn_time': round(self.nn_time, 3),
            # 'decoder_time': round(self.decoder_time, 3),
            # 'n_images': self.n_images,
        }
        if additional_data:
            data = dict(**data, **additional_data)
        with open(out_name, 'w') as f:
            json.dump(data, f)
        LOG.info('wrote %s', out_name)

    # def _annotation_name_2017(self):
    #     lookup = {}
    #     for fn in os.listdir(self.gt_annotations_2017_folder):
    #         if not fn.endswith('.json'):
    #             continue
    #         sequence_index, _, __ = fn.partition('_')
    #         sequence_index = int(sequence_index)
    #         lookup[sequence_index] = fn

    #     current_sequence_index, _, __ = os.path.basename(self.annotation_file).partition('_')
    #     current_sequence_index = int(current_sequence_index)

    #     return lookup.get(current_sequence_index)

    # def _image_names_2017(self, annotation_file):
    #     with open(annotation_file, 'r') as f:
    #         data_2017 = json.load(f)
    #     return [i['image'][0]['name'] for i in data_2017['annolist']]

    @staticmethod
    def _format_2017(image_names, predictions):
        r"""Create a datastructure that transforms to Posetrack2017 json.

        2017 target:
{
   "annolist": [
       {
           "image": [
               {
                  "name": "images\/bonn_5sec\/000342_mpii\/00000001.jpg"
               }
           ],
           "annorect": [
               {
                   "x1": [625],
                   "y1": [94],
                   "x2": [681],
                   "y2": [178],
                   "score": [0.9],
                   "track_id": [0],
                   "annopoints": [
                       {
                           "point": [
                               {
                                   "id": [0],
                                   "x": [394],
                                   "y": [173],
                                   "score": [0.7],
                               },
                               { ... }
                           ]
                       }
                   ]
                },
                { ... }
           ],
       },
       { ... }
   ]
}

        2018 source in self.data:
    [
        {
            "bbox_head": [
                625,
                94,
                56,
                84
            ],
            "keypoints": [
                659,
                152,
                ...
            ],
            "track_id": 0,
            "image_id": 10003420000,
            "bbox": [
                331.64999999999998,
                105.21967275,
                427.70000000000005,
                154.62123949999997
            ],
            "scores": [],
            "category_id": 1,
            "id": 1000342000000
        },
        ...
    ]

        images contains:
        {
            "has_no_densepose": true,
            "is_labeled": false,
            "file_name": "images/val/000342_mpii_test/000093.jpg",
            "nframes": 100,
            "frame_id": 10003420093,
            "vid_id": "000342",
            "id": 10003420093
        },
        """

        annorect_by_image = defaultdict(list)
        for ann in predictions:
            kps2018 = np.array(ann['keypoints']).reshape(-1, 3)
            kps2017 = kps2018[POSETRACK_INDEX2017TO2018]

            annorect = {
                'x1': [ann['bbox'][0]],
                'y1': [ann['bbox'][1]],
                'x2': [ann['bbox'][2]],
                'y2': [ann['bbox'][3]],
                'score': [ann['score']],
                'track_id': [ann['track_id']],
                'annopoints': [
                    {
                        'point': [
                            {
                                'id': [kpi],
                                'x': [xyv[0]],
                                'y': [xyv[1]],
                                'score': [xyv[2]],
                            }
                            for kpi, xyv in enumerate(kps2017)
                            if xyv[2] > 0.0
                        ]
                    }
                ]
            }
            annorect_by_image[ann['image_file']].append(annorect)

        # image_ids = sorted(annorect_by_image.keys())
        # print(len(image_ids), len(image_names))
        # assert len(image_ids) == len(image_names)
        return [
            {
                'image': {
                    'name': image_name,
                },
                'annorect': annorect_by_image.get(image_name, []),
            }
            for image_name in image_names
        ]

    def write_predictions(self, filename, *, additional_data=None):
        output_dir = '{}-{}'.format(filename, self.output_dir_suffix)
        for annotation_file in self.predictions_by_file.keys():
            if self.output_format == '2018':
                self._write2018(output_dir, annotation_file, additional_data=additional_data)
            elif self.output_format == '2017':
                self._write2017(output_dir, annotation_file, additional_data=additional_data)
            else:
                raise NotImplementedError

        if self.ground_truth_directory is None:
            return

        # use poseval to evaluate right away
        gt_dir = os.path.dirname(self.ground_truth_directory)
        if not gt_dir.endswith('/'):
            gt_dir = gt_dir + '/'

        pred_dir = output_dir
        if not pred_dir.endswith('/'):
            pred_dir = pred_dir + '/'

        out_dir = output_dir
        if out_dir.endswith('/'):
            out_dir = out_dir[:-1]
        out_dir = out_dir + '-poseval/'

        cmd = [
            'python', '-m', 'poseval.evaluate',
            '--groundTruth', gt_dir,
            '--predictions', pred_dir,
            '--outputDir', out_dir,
            '--evalPoseTracking',
            '--evalPoseEstimation',
            '--saveEvalPerSequence',
        ]
        LOG.info('eval command: %s', ' '.join(cmd))
        subprocess.run(cmd, check=True)
        self.print_by_sequence(out_dir)

        self._written_mot_stats_file = os.path.join(out_dir, 'total_MOT_metrics.json')
        self._written_ap_stats_file = os.path.join(out_dir, 'total_AP_metrics.json')

    @staticmethod
    def print_by_sequence(out_dir):
        mot_files = glob.glob(os.path.join(out_dir, '*_MOT_metrics.json'))

        mota = {}
        for file_name in mot_files:
            with open(file_name, 'r') as f:
                d = json.load(f)
            identifier = os.path.basename(file_name).replace('_MOT_metrics.json', '')
            mota[identifier] = d.get('mota', [-1.0])[-1]

        print('sequence, mota')
        for sequence, m in sorted(mota.items(), key=lambda x: x[1]):
            print(sequence, m)



import logging

from .base import Base

LOG = logging.getLogger(__name__)


class CifHr(Base):
    """Visualize the CifHr map."""

    def __init__(self, *, stride=1, field_names=None):
        super().__init__('cif')

        self.stride = stride
        self.field_names = field_names

    def predicted(self, fields, low=0.0):
        for f in self.indices('hr'):
            LOG.debug('%d (field name: %s)',
                      f, self.field_names[f] if self.field_names else 'unknown')
            with self.image_canvas(self.processed_image(), margin=[0.0, 0.01, 0.05, 0.01]) as ax:
                o = ax.imshow(fields[f], alpha=0.9, vmin=low, vmax=low + 1.0, cmap='Oranges')
                self.colorbar(ax, o)



"""Datasets and tools to load data in batches."""

from .collate import (
    collate_images_anns_meta,
    collate_images_targets_meta,
    collate_tracking_images_targets_meta,
)
from .factory import cli, configure, factory, DATAMODULES
from .image_list import ImageList, NumpyImageList, PilImageList
from .loader_with_reset import LoaderWithReset
from .module import DataModule
from .multimodule import MultiDataModule
from .torch_dataset import TorchDataset



import copy
import logging

import PIL
import torch

from ..preprocess import Preprocess

LOG = logging.getLogger(__name__)


class CameraShift(Preprocess):
    def __init__(self, max_shift=100):
        super().__init__()

        self.max_shift = max_shift

    def __call__(self, image, anns, meta):
        meta = copy.deepcopy(meta)
        anns = copy.deepcopy(anns)

        xy_shift = (torch.rand(2).numpy() - 0.5) * 2 * self.max_shift
        xy_shift *= meta.get('group_i', 1.0)

        # shift image
        affine_params = (1.0, 0.0, xy_shift[0], 0.0, 1.0, xy_shift[1])
        image = image.transform(image.size, PIL.Image.AFFINE, affine_params,
                                fillcolor=(127, 127, 127))

        # shift all annotations
        for ann in anns:
            ann['keypoints'][:, :2] += xy_shift
            ann['bbox'][:2] += xy_shift

        # adjust meta
        LOG.debug('meta before: %s', meta)
        meta['offset'] += xy_shift
        meta['valid_area'][:2] += xy_shift
        LOG.debug('meta after: %s', meta)

        return image, anns, meta



import logging
import json
import zipfile
import copy

import numpy as np
try:
    from xtcocotools.cocoeval import COCOeval
except ImportError:
    pass
from openpifpaf.metric.base import Base

try:
    import pycocotools.coco
except ImportError:
    pass

LOG = logging.getLogger(__name__)


class WholebodyMetric(Base):
    text_labels_keypoints = ['AP', 'AP0.5', 'AP0.75', 'APM', 'APL',
                             'AR', 'AR0.5', 'AR0.75', 'ARM', 'ARL']
    text_labels_bbox = ['AP', 'AP0.5', 'AP0.75', 'APS', 'APM', 'APL',
                        'ART1', 'ART10', 'AR', 'ARS', 'ARM', 'ARL']

    def __init__(self, coco, *,
                 max_per_image=20,
                 category_ids=None,
                 iou_type='keypoints',
                 small_threshold=0.0,
                 keypoint_oks_sigmas=None):
        super().__init__()

        if category_ids is None:
            category_ids = [1]

        self.max_per_image = max_per_image
        self.category_ids = category_ids
        self.iou_type = iou_type
        self.small_threshold = small_threshold
        self.keypoint_oks_sigmas = keypoint_oks_sigmas

        self.predictions = []
        self.image_ids = []
        self.eval = None
        self.coco = pycocotools.coco.COCO(coco)

        self.sigmas_body = [0.026, 0.025, 0.025, 0.035, 0.035, 0.079, 0.079, 0.072, 0.072,
                            0.062, 0.062, 0.107, 0.107, 0.087,
                            0.087, 0.089, 0.089]
        self.sigmas_foot = [0.068, 0.066, 0.066, 0.092, 0.094, 0.094]
        self.sigmas_face = [0.042, 0.043, 0.044, 0.043, 0.040, 0.035, 0.031, 0.025, 0.020, 0.023,
                            0.029, 0.032, 0.037, 0.038, 0.043,
                            0.041, 0.045, 0.013, 0.012, 0.011, 0.011, 0.012, 0.012, 0.011, 0.011,
                            0.013, 0.015, 0.009, 0.007, 0.007,
                            0.007, 0.012, 0.009, 0.008, 0.016, 0.010, 0.017, 0.011, 0.009, 0.011,
                            0.009, 0.007, 0.013, 0.008, 0.011,
                            0.012, 0.010, 0.034, 0.008, 0.008, 0.009, 0.008, 0.008, 0.007, 0.010,
                            0.008, 0.009, 0.009, 0.009, 0.007,
                            0.007, 0.008, 0.011, 0.008, 0.008, 0.008, 0.01, 0.008]
        self.sigmas_lefthand = [0.029, 0.022, 0.035, 0.037, 0.047, 0.026, 0.025, 0.024,
                                0.035, 0.018, 0.024, 0.022, 0.026, 0.017,
                                0.021, 0.021, 0.032, 0.02, 0.019, 0.022, 0.031]
        self.sigmas_righthand = [0.029, 0.022, 0.035, 0.037, 0.047, 0.026, 0.025, 0.024,
                                 0.035, 0.018, 0.024, 0.022, 0.026, 0.017,
                                 0.021, 0.021, 0.032, 0.02, 0.019, 0.022, 0.031]

        self.sigmas_wholebody = self.sigmas_body + self.sigmas_foot + self.sigmas_face + \
            self.sigmas_lefthand + self.sigmas_righthand

        if self.iou_type == 'keypoints':
            self.text_labels = self.text_labels_keypoints
        elif self.iou_type == 'bbox':
            self.text_labels = self.text_labels_bbox
        else:
            LOG.warning('Unknown iou type "%s". Specify text_labels yourself.', self.iou_type)

        LOG.debug('max = %d, category ids = %s, iou_type = %s',
                  self.max_per_image, self.category_ids, self.iou_type)

    def _stats(self, predictions=None, image_ids=None):
        # from pycocotools.cocoeval import COCOeval
        if predictions is None:
            predictions = self.predictions
        if image_ids is None:
            image_ids = self.image_ids

        coco_eval = self.coco.loadRes(predictions)

        for count in coco_eval.anns:
            ann_orig = copy.deepcopy(coco_eval.anns[count])
            coco_eval.anns[count]["lefthand_kpts"] = ann_orig["keypoints"][91 * 3:112 * 3]
            coco_eval.anns[count]["righthand_kpts"] = ann_orig["keypoints"][112 * 3:133 * 3]
            coco_eval.anns[count]["face_kpts"] = ann_orig["keypoints"][23 * 3:91 * 3]
            coco_eval.anns[count]["foot_kpts"] = ann_orig["keypoints"][17 * 3:23 * 3]
            coco_eval.anns[count]["keypoints"] = ann_orig["keypoints"][0:17 * 3]

        coco_eval.anno_file = coco_eval.anns  # xtpycocotools style
        self.coco.anno_file = self.coco.anns
        eval_body = COCOeval(self.coco, coco_eval, 'keypoints_body', np.array(self.sigmas_body),
                             use_area=True)
        eval_body.evaluate()
        eval_body.accumulate()
        eval_body.summarize()
        eval_foot = COCOeval(self.coco, coco_eval, 'keypoints_foot', np.array(self.sigmas_foot),
                             use_area=True)
        eval_foot.evaluate()
        eval_foot.accumulate()
        eval_foot.summarize()
        eval_face = COCOeval(self.coco, coco_eval, 'keypoints_face', np.array(self.sigmas_face),
                             use_area=True)
        eval_face.evaluate()
        eval_face.accumulate()
        eval_face.summarize()
        eval_lh = COCOeval(self.coco, coco_eval, 'keypoints_lefthand',
                           np.array(self.sigmas_lefthand), use_area=True)
        eval_lh.evaluate()
        eval_lh.accumulate()
        eval_lh.summarize()
        eval_rh = COCOeval(self.coco, coco_eval, 'keypoints_righthand',
                           np.array(self.sigmas_righthand), use_area=True)
        eval_rh.evaluate()
        eval_rh.accumulate()
        eval_rh.summarize()
        eval_wb = COCOeval(self.coco, coco_eval, 'keypoints_wholebody',
                           np.array(self.sigmas_wholebody), use_area=True)
        eval_wb.evaluate()
        eval_wb.accumulate()
        eval_wb.summarize()
        return eval_wb.stats

    # pylint: disable=unused-argument
    def accumulate(self, predictions, image_meta, *, ground_truth=None):
        image_id = int(image_meta['image_id'])
        self.image_ids.append(image_id)

        if self.small_threshold:
            predictions = [pred for pred in predictions
                           if pred.scale(v_th=0.01) >= self.small_threshold]
        if len(predictions) > self.max_per_image:
            predictions = predictions[:self.max_per_image]

        image_annotations = []
        for pred in predictions:
            pred_data = pred.json_data()
            pred_data['image_id'] = image_id
            pred_data = {
                k: v for k, v in pred_data.items()
                if k in ('category_id', 'score', 'keypoints', 'bbox', 'image_id')
            }
            image_annotations.append(pred_data)

        # force at least one annotation per image (for pycocotools)
        if not image_annotations:
            n_keypoints = 133
            image_annotations.append({
                'image_id': image_id,
                'category_id': 1,
                'keypoints': np.zeros((n_keypoints * 3,)).tolist(),
                'bbox': [0, 0, 1, 1],
                'score': 0.001,
            })

        if LOG.getEffectiveLevel() == logging.DEBUG:
            self._stats(image_annotations, [image_id])
            LOG.debug(image_meta)

        self.predictions += image_annotations

    def write_predictions(self, filename, *, additional_data=None):
        predictions = [
            {k: v for k, v in annotation.items()
             if k in ('image_id', 'category_id', 'keypoints', 'score')}
            for annotation in self.predictions
        ]
        predictions_wb = copy.deepcopy(predictions)
        for ann in predictions_wb:
            ann_orig = copy.deepcopy(ann)
            ann["lefthand_kpts"] = ann_orig["keypoints"][91 * 3:112 * 3]
            ann["righthand_kpts"] = ann_orig["keypoints"][112 * 3:133 * 3]
            ann["face_kpts"] = ann_orig["keypoints"][23 * 3:91 * 3]
            ann["foot_kpts"] = ann_orig["keypoints"][17 * 3:23 * 3]
            ann["keypoints"] = ann_orig["keypoints"][0:17 * 3]
        with open(filename + '.pred_wb.json', 'w') as f:
            json.dump(predictions_wb, f)
        LOG.info('wrote %s.pred_wb.json', filename)
        with open(filename + '.pred.json', 'w') as f:
            json.dump(predictions, f)
        LOG.info('wrote %s.pred.json', filename)
        with zipfile.ZipFile(filename + '.zip', 'w') as myzip:
            myzip.write(filename + '.pred.json', arcname='predictions.json')
        LOG.info('wrote %s.zip', filename)

        if additional_data:
            with open(filename + '.pred_meta.json', 'w') as f:
                json.dump(additional_data, f)
            LOG.info('wrote %s.pred_meta.json', filename)

    def stats(self):
        data = {
            'stats': self._stats().tolist(),
            'text_labels': self.text_labels,
        }

        return data



from collections import defaultdict
import logging
import time
from typing import List

import torch
import torchvision

from .decoder import Decoder
from ..annotation import AnnotationDet
from .. import headmeta, visualizer

LOG = logging.getLogger(__name__)


class CifDet(Decoder):
    iou_threshold = 0.5
    instance_threshold = 0.15
    nms_by_category = True
    occupancy_visualizer = visualizer.Occupancy()
    suppression = 0.1

    def __init__(self, head_metas: List[headmeta.CifDet], *, visualizers=None):
        super().__init__()
        self.metas = head_metas

        # prefer decoders with more classes
        self.priority = -1.0  # prefer keypoints over detections
        self.priority += sum(m.n_fields for m in head_metas) / 1000.0

        self.visualizers = visualizers
        if self.visualizers is None:
            self.visualizers = [visualizer.CifDet(meta) for meta in self.metas]

        self.cpp_decoder = torch.classes.openpifpaf_decoder.CifDet()

        self.timers = defaultdict(float)

    @classmethod
    def factory(cls, head_metas):
        # TODO: multi-scale
        return [
            CifDet([meta])
            for meta in head_metas
            if isinstance(meta, headmeta.CifDet)
        ]

    def __call__(self, fields):
        start = time.perf_counter()

        if self.visualizers:
            for vis, meta in zip(self.visualizers, self.metas):
                vis.predicted(fields[meta.head_index])

        categories, scores, boxes = self.cpp_decoder.call(
            fields[self.metas[0].head_index],
            self.metas[0].stride,
        )

        # nms
        if self.nms_by_category:
            keep_index = torchvision.ops.batched_nms(boxes, scores, categories, self.iou_threshold)
        else:
            keep_index = torchvision.ops.nms(boxes, scores, self.iou_threshold)
        pre_nms_scores = scores.clone()
        scores *= self.suppression
        scores[keep_index] = pre_nms_scores[keep_index]
        filter_mask = scores > self.instance_threshold
        categories = categories[filter_mask]
        scores = scores[filter_mask]
        boxes = boxes[filter_mask]
        LOG.debug('cpp annotations = %d (%.1fms)',
                  len(scores),
                  (time.perf_counter() - start) * 1000.0)

        # convert to py
        annotations_py = []
        boxes_np = boxes.numpy()
        boxes_np[:, 2:] -= boxes_np[:, :2]  # convert to xywh
        for category, score, box in zip(categories, scores, boxes_np):
            ann = AnnotationDet(self.metas[0].categories)
            ann.set(int(category), float(score), box)
            annotations_py.append(ann)

        LOG.info('annotations %d, decoder = %.1fms',
                 len(annotations_py),
                 (time.perf_counter() - start) * 1000.0)
        return annotations_py




import os

import numpy as np

_CATEGORIES = ['cat', 'cow', 'dog', 'sheep', 'horse']  # only for preprocessing

ANIMAL_KEYPOINTS = [
    'Nose',         # 1
    'L_eye',        # 2
    'R_eye',        # 3
    'L_ear',        # 4
    'R_ear',        # 5
    'Throat',       # 6
    'Tail',         # 7
    'withers',      # 8
    'L_F_elbow',    # 9
    'R_F_elbow',    # 10
    'L_B_elbow',    # 11
    'R_B_elbow',    # 12
    'L_F_knee',     # 13
    'R_F_knee',     # 14
    'L_B_knee',     # 15
    'R_B_knee',     # 16
    'L_F_paw',      # 17
    'R_F_paw',      # 18
    'L_B_paw',      # 19
    'R_B_paw',      # 20
]


HFLIP = {
    'L_eye': 'R_eye',
    'R_eye': 'L_eye',
    'L_ear': 'R_ear',
    'R_ear': 'L_ear',
    'L_F_elbow': 'R_F_elbow',
    'R_F_elbow': 'L_F_elbow',
    'L_B_elbow': 'R_B_elbow',
    'R_B_elbow': 'L_B_elbow',
    'L_F_knee': 'R_F_knee',
    'R_F_knee': 'L_F_knee',
    'L_B_knee': 'R_B_knee',
    'R_B_knee': 'L_B_knee',
    'L_F_paw': 'R_F_paw',
    'R_F_paw': 'L_F_paw',
    'L_B_paw': 'R_B_paw',
    'R_B_paw': 'L_B_paw',
}


ALTERNATIVE_NAMES = [
    'Nose',         # 1
    'L_Eye',        # 2
    'R_Eye',        # 3
    'L_EarBase',    # 4
    'R_EarBase',    # 5
    'Throat',       # 6
    'TailBase',     # 7
    'Withers',      # 8
    'L_F_Elbow',    # 9
    'R_F_Elbow',    # 10
    'L_B_Elbow',    # 11
    'R_B_Elbow',    # 12
    'L_F_Knee',     # 13
    'R_F_Knee',     # 14
    'L_B_Knee',     # 15
    'R_B_Knee',     # 16
    'L_F_Paw',      # 17
    'R_F_Paw',      # 18
    'L_B_Paw',      # 19
    'R_B_Paw',      # 20
]


ANIMAL_SKELETON = [
    (1, 2), (1, 3), (1, 6), (2, 4), (3, 5), (2, 3), (6, 8), (6, 9), (6, 10), (7, 8),
    (7, 11), (7, 12),  # Torso + Face
    (10, 14), (14, 18), (9, 13), (13, 17), (12, 16), (16, 20), (11, 15), (15, 19)  # Legs
]


ANIMAL_SIGMAS = [
    0.026,  # nose
    0.025,  # eyes
    0.025,  # eyes
    0.035,  # ears
    0.035,  # ears
    0.079,  # throat
    0.079,  # tail
    0.079,  # withers
    0.072,  # elbows
    0.072,  # elbows
    0.072,  # elbows
    0.072,  # elbows
    0.087,  # knees
    0.087,  # knees
    0.087,  # knees
    0.087,  # knees
    0.089,  # ankles
    0.089,  # ankles
    0.089,  # ankles
    0.089,  # ankles
]

split, error = divmod(len(ANIMAL_KEYPOINTS), 4)
ANIMAL_SCORE_WEIGHTS = [5.0] * split + [3.0] * split + [1.0] * split + [0.5] * split + [0.1] * error

ANIMAL_CATEGORIES = ['animal']

ANIMAL_POSE = np.array([
    [0.0, 4.3, 2.0],  # 'nose',            # 1
    [-0.4, 4.7, 2.0],  # 'left_eye',        # 2
    [0.4, 4.7, 2.0],  # 'right_eye',       # 3
    [-0.7, 5.0, 2.0],  # 'left_ear',        # 4
    [0.7, 5.0, 2.0],  # 'right_ear',       # 5
    [0.2, 3.0, 2.0],  # 'throat',            # 6
    [6.7, 3.8, 2.0],  # 'tail',             # 7
    [0.8, 4.0, 2.0],  # 'withers',         # 8
    [1.0, 2.0, 2.0],  # 'L_F_elbow',      # 9
    [0.6, 2.2, 2.0],  # 'R_F_elbow',     # 10
    [5.8, 2.1, 2.0],  # 'L_B_elbow',      # 11
    [6.3, 2.3, 2.0],  # 'R_B_elbow',     # 12
    [0.8, 0.8, 2.0],  # 'L_F_Knee',     # 13
    [0.4, 1.0, 2.0],  # 'R_F_Knee',     # 14
    [6.0, 0.9, 2.0],  # 'L_B_Knee',     # 15
    [6.5, 1.1, 2.0],  # 'R_B_Knee',     # 16
    [1.0, 0.0, 2.0],  # 'L_F_Paw',      # 17
    [0.6, 0.2, 2.0],  # 'R_F_Paw',      # 18
    [6.0, 0.1, 2.0],  # 'L_B_Paw',      # 19
    [6.5, 0.3, 2.0],  # 'R_B_Paw',      # 20
])


assert len(ANIMAL_POSE) == len(ANIMAL_KEYPOINTS) == len(ALTERNATIVE_NAMES) == len(ANIMAL_SIGMAS) \
       == len(ANIMAL_SCORE_WEIGHTS), "dimensions!"


def draw_ann(ann, *, keypoint_painter, filename=None, margin=0.5, aspect=None, **kwargs):
    from openpifpaf import show  # pylint: disable=import-outside-toplevel

    bbox = ann.bbox()
    xlim = bbox[0] - margin, bbox[0] + bbox[2] + margin
    ylim = bbox[1] - margin, bbox[1] + bbox[3] + margin
    if aspect == 'equal':
        fig_w = 5.0
    else:
        fig_w = 5.0 / (ylim[1] - ylim[0]) * (xlim[1] - xlim[0])

    with show.canvas(filename, figsize=(fig_w, 5), nomargin=True, **kwargs) as ax:
        ax.set_axis_off()
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)

        if aspect is not None:
            ax.set_aspect(aspect)

        keypoint_painter.annotation(ax, ann)


def draw_skeletons(pose):
    from openpifpaf.annotation import Annotation  # pylint: disable=import-outside-toplevel
    from openpifpaf import show  # pylint: disable=import-outside-toplevel

    scale = np.sqrt(
        (np.max(pose[:, 0]) - np.min(pose[:, 0]))
        * (np.max(pose[:, 1]) - np.min(pose[:, 1]))
    )

    show.KeypointPainter.show_joint_scales = True
    show.KeypointPainter.font_size = 0
    keypoint_painter = show.KeypointPainter()

    ann = Annotation(
        keypoints=ANIMAL_KEYPOINTS, skeleton=ANIMAL_SKELETON, score_weights=ANIMAL_SCORE_WEIGHTS)
    ann.set(pose, np.array(ANIMAL_SIGMAS) * scale)
    os.makedirs('all-images', exist_ok=True)
    draw_ann(ann, filename='all-images/skeleton_animal.png', keypoint_painter=keypoint_painter)


def print_associations():
    for j1, j2 in ANIMAL_SKELETON:
        print(ANIMAL_SKELETON[j1 - 1], '-', ANIMAL_KEYPOINTS[j2 - 1])


if __name__ == '__main__':
    print_associations()
    draw_skeletons(ANIMAL_POSE)



from contextlib import contextmanager
import logging
import os

import numpy as np

try:
    import matplotlib.pyplot as plt  # pylint: disable=import-error
except ModuleNotFoundError as err:
    if err.name != 'matplotlib':
        raise err
    plt = None


LOG = logging.getLogger(__name__)


class Canvas:
    """Canvas for plotting.

    All methods expose Axes objects. To get Figure objects, you can ask the axis
    `ax.get_figure()`.
    """

    all_images_directory = None
    all_images_count = 0
    show = False
    image_width = 7.0
    image_height = None
    blank_dpi = 200
    image_dpi_factor = 2.0
    image_min_dpi = 50.0
    out_file_extension = 'jpeg'
    white_overlay = False

    @classmethod
    def generic_name(cls):
        if cls.all_images_directory is None:
            return None

        os.makedirs(cls.all_images_directory, exist_ok=True)

        cls.all_images_count += 1
        return os.path.join(cls.all_images_directory,
                            '{:04}.{}'.format(cls.all_images_count, cls.out_file_extension))

    @classmethod
    @contextmanager
    def blank(cls, fig_file=None, *, dpi=None, nomargin=False, **kwargs):
        if plt is None:
            raise Exception('please install matplotlib')
        if fig_file is None:
            fig_file = cls.generic_name()
        if dpi is None:
            dpi = cls.blank_dpi

        if 'figsize' not in kwargs:
            kwargs['figsize'] = (10, 6)

        if nomargin:
            if 'gridspec_kw' not in kwargs:
                kwargs['gridspec_kw'] = {}
            kwargs['gridspec_kw']['wspace'] = 0
            kwargs['gridspec_kw']['hspace'] = 0
            kwargs['gridspec_kw']['left'] = 0.0
            kwargs['gridspec_kw']['right'] = 1.0
            kwargs['gridspec_kw']['top'] = 1.0
            kwargs['gridspec_kw']['bottom'] = 0.0
        fig, ax = plt.subplots(dpi=dpi, **kwargs)

        yield ax

        fig.set_tight_layout(not nomargin)
        if fig_file:
            LOG.debug('writing image to %s', fig_file)
            fig.savefig(fig_file)
        if cls.show:
            plt.show()
        plt.close(fig)

    @classmethod
    @contextmanager
    def image(cls, image, fig_file=None, *, margin=None, **kwargs):
        if plt is None:
            raise Exception('please install matplotlib')
        if fig_file is None:
            fig_file = cls.generic_name()

        image = np.asarray(image)

        if margin is None:
            margin = [0.0, 0.0, 0.0, 0.0]
        elif isinstance(margin, float):
            margin = [margin, margin, margin, margin]
        assert len(margin) == 4

        if 'figsize' not in kwargs:
            # compute figure size: use image ratio and take the drawable area
            # into account that is left after subtracting margins.
            image_ratio = image.shape[0] / image.shape[1]
            image_area_ratio = (1.0 - margin[1] - margin[3]) / (1.0 - margin[0] - margin[2])
            if cls.image_width is not None:
                kwargs['figsize'] = (
                    cls.image_width,
                    cls.image_width * image_ratio / image_area_ratio
                )
            elif cls.image_height:
                kwargs['figsize'] = (
                    cls.image_height * image_area_ratio / image_ratio,
                    cls.image_height
                )

        dpi = max(cls.image_min_dpi, image.shape[1] / kwargs['figsize'][0] * cls.image_dpi_factor)
        fig = plt.figure(dpi=dpi, **kwargs)
        ax = plt.Axes(fig, [0.0 + margin[0],
                            0.0 + margin[1],
                            1.0 - margin[2],
                            1.0 - margin[3]])
        ax.set_axis_off()
        ax.set_xlim(-0.5, image.shape[1] - 0.5)  # imshow uses center-pixel-coordinates
        ax.set_ylim(image.shape[0] - 0.5, -0.5)
        fig.add_axes(ax)
        ax.imshow(image)
        if cls.white_overlay:
            white_screen(ax, cls.white_overlay)

        yield ax

        if fig_file:
            LOG.debug('writing image to %s', fig_file)
            fig.savefig(fig_file)
        if cls.show:
            plt.show()
        plt.close(fig)

    @classmethod
    @contextmanager
    def annotation(cls, ann, *,
                   filename=None,
                   margin=0.5,
                   fig_w=None,
                   fig_h=5.0,
                   **kwargs):
        bbox = ann.bbox()
        xlim = bbox[0] - margin, bbox[0] + bbox[2] + margin
        ylim = bbox[1] - margin, bbox[1] + bbox[3] + margin
        if fig_w is None:
            fig_w = fig_h / (ylim[1] - ylim[0]) * (xlim[1] - xlim[0])
        fig_w *= kwargs.get('ncols', 1)
        fig_h *= kwargs.get('nrows', 1)

        with cls.blank(filename, figsize=(fig_w, fig_h), nomargin=True, **kwargs) as ax:
            iter_ax = [ax] if hasattr(ax, 'set_axis_off') else ax
            for ax_ in iter_ax:
                ax_.set_axis_off()
                ax_.set_xlim(*xlim)
                ax_.set_ylim(*ylim)

            yield ax


# keep previous interface for now:
canvas = Canvas.blank
image_canvas = Canvas.image
annotation_canvas = Canvas.annotation


def white_screen(ax, alpha=0.9):
    ax.add_patch(
        plt.Rectangle((0, 0), 1, 1, transform=ax.transAxes, alpha=alpha,
                      facecolor='white')
    )



import logging

import numpy as np

from ..configurable import Configurable

try:
    import matplotlib
    import matplotlib.animation
    import matplotlib.collections
    import matplotlib.patches
except ImportError:
    matplotlib = None


LOG = logging.getLogger(__name__)


class DetectionPainter:
    def __init__(self, *, xy_scale=1.0):
        self.xy_scale = xy_scale

    def annotation(self, ax, ann, *, color=None, text=None, subtext=None):
        if color is None:
            color = 0
        if isinstance(color, (int, np.integer)):
            color = matplotlib.cm.get_cmap('tab20')((color % 20 + 0.05) / 20)

        if text is None:
            text = ann.category
            if getattr(ann, 'id_', None):
                text += ' ({})'.format(ann.id_)
        if subtext is None and ann.score:
            subtext = '{:.0%}'.format(ann.score)

        x, y, w, h = ann.bbox * self.xy_scale
        if w < 5.0:
            x -= 2.0
            w += 4.0
        if h < 5.0:
            y -= 2.0
            h += 4.0

        # draw box
        ax.add_patch(
            matplotlib.patches.Rectangle(
                (x, y), w, h, fill=False, color=color, linewidth=1.0))

        # draw text
        ax.annotate(
            text,
            (x, y),
            fontsize=8,
            xytext=(5.0, 5.0),
            textcoords='offset points',
            color='white', bbox={'facecolor': color, 'alpha': 0.5, 'linewidth': 0},
        )
        if subtext is not None:
            ax.annotate(
                subtext,
                (x, y),
                fontsize=5,
                xytext=(5.0, 18.0 + 3.0),
                textcoords='offset points',
                color='white', bbox={'facecolor': color, 'alpha': 0.5, 'linewidth': 0},
            )


class CrowdPainter:
    def __init__(self, *, xy_scale=1.0):
        self.xy_scale = xy_scale

    @staticmethod
    def draw_polygon(ax, outlines, *, alpha=0.5, color='orange'):
        for outline in outlines:
            assert outline.shape[1] == 2

        patches = []
        for outline in outlines:
            polygon = matplotlib.patches.Polygon(
                outline[:, :2], color=color, facecolor=color, alpha=alpha)
            patches.append(polygon)
        ax.add_collection(matplotlib.collections.PatchCollection(patches, match_original=True))

    def annotation(self, ax, ann, *, color=None, text=None, subtext=None):
        if color is None:
            color = 0
        if isinstance(color, (int, np.integer)):
            color = matplotlib.cm.get_cmap('tab20')((color % 20 + 0.05) / 20)

        if text is None:
            text = '{} (crowd)'.format(ann.category)
            if getattr(ann, 'id_', None):
                text += ' ({})'.format(ann.id_)

        x, y, w, h = ann.bbox * self.xy_scale
        if w < 5.0:
            x -= 2.0
            w += 4.0
        if h < 5.0:
            y -= 2.0
            h += 4.0

        # draw box
        ax.add_patch(
            matplotlib.patches.Rectangle(
                (x, y), w, h, fill=False, color=color, linewidth=1.0, linestyle='dotted'))

        # draw text
        ax.annotate(
            text,
            (x, y),
            fontsize=8,
            xytext=(5.0, 5.0),
            textcoords='offset points',
            color='white', bbox={'facecolor': color, 'alpha': 0.5, 'linewidth': 0},
        )
        if subtext is not None:
            ax.annotate(
                subtext,
                (x, y),
                fontsize=5,
                xytext=(5.0, 18.0 + 3.0),
                textcoords='offset points',
                color='white', bbox={'facecolor': color, 'alpha': 0.5, 'linewidth': 0},
            )


class KeypointPainter(Configurable):
    """Paint poses.

    The constructor can take any class attribute as parameter and
    overwrite the global default for that instance.

    Example to create a KeypointPainter with thick lines:

    >>> kp = KeypointPainter(line_width=48)

    """

    show_box = False
    show_joint_confidences = False
    show_joint_scales = False
    show_decoding_order = False
    show_frontier_order = False
    show_only_decoded_connections = False

    textbox_alpha = 0.5
    text_color = 'white'
    monocolor_connections = False
    line_width = None
    marker_size = None
    solid_threshold = 0.5
    font_size = 8

    def __init__(self, *,
                 xy_scale=1.0,
                 highlight=None,
                 highlight_invisible=False,
                 **kwargs):
        super().__init__(**kwargs)

        self.xy_scale = xy_scale
        self.highlight = highlight
        self.highlight_invisible = highlight_invisible

        # set defaults for line_width and marker_size depending on monocolor
        if self.line_width is None:
            self.line_width = 2 if self.monocolor_connections else 6
        if self.marker_size is None:
            if self.monocolor_connections:
                self.marker_size = max(self.line_width + 1, int(self.line_width * 3.0))
            else:
                self.marker_size = max(1, int(self.line_width * 0.5))

        LOG.debug('color connections = %s, lw = %d, marker = %d',
                  self.monocolor_connections, self.line_width, self.marker_size)

    def _draw_skeleton(self, ax, x, y, v, *,
                       skeleton, skeleton_mask=None, color=None, alpha=1.0, **kwargs):
        if not np.any(v > 0):
            return

        if skeleton_mask is None:
            skeleton_mask = [True for _ in skeleton]
        assert len(skeleton) == len(skeleton_mask)

        # connections
        lines, line_colors, line_styles = [], [], []
        for ci, ((j1i, j2i), mask) in enumerate(zip(np.array(skeleton) - 1, skeleton_mask)):
            if not mask:
                continue
            c = color
            if not self.monocolor_connections:
                c = matplotlib.cm.get_cmap('tab20')((ci % 20 + 0.05) / 20)
            if v[j1i] > 0 and v[j2i] > 0:
                lines.append([(x[j1i], y[j1i]), (x[j2i], y[j2i])])
                line_colors.append(c)
                if v[j1i] > self.solid_threshold and v[j2i] > self.solid_threshold:
                    line_styles.append('solid')
                else:
                    line_styles.append('dashed')
        ax.add_collection(matplotlib.collections.LineCollection(
            lines, colors=line_colors,
            linewidths=kwargs.get('linewidth', self.line_width),
            linestyles=kwargs.get('linestyle', line_styles),
            capstyle='round',
            alpha=alpha,
        ))

        # joints
        ax.scatter(
            x[v > 0.0], y[v > 0.0], s=self.marker_size**2, marker='.',
            color=color if self.monocolor_connections else 'white',
            edgecolor='k' if self.highlight_invisible else None,
            zorder=2,
            alpha=alpha,
        )

        # highlight joints
        if self.highlight is not None:
            highlight_v = np.zeros_like(v)
            highlight_v[self.highlight] = 1
            highlight_v = np.logical_and(v, highlight_v)

            ax.scatter(
                x[highlight_v], y[highlight_v], s=self.marker_size**2, marker='.',
                color=color if self.monocolor_connections else 'white',
                edgecolor='k' if self.highlight_invisible else None,
                zorder=2,
                alpha=alpha,
            )

    def keypoints(self, ax, keypoint_sets, *,
                  skeleton, scores=None, color=None, colors=None, texts=None):
        if keypoint_sets is None:
            return

        if color is None and colors is None:
            colors = range(len(keypoint_sets))

        for i, kps in enumerate(np.asarray(keypoint_sets)):
            assert kps.shape[1] == 3
            x = kps[:, 0] * self.xy_scale
            y = kps[:, 1] * self.xy_scale
            v = kps[:, 2]

            if colors is not None:
                color = colors[i]

            if isinstance(color, (int, np.integer)):
                color = matplotlib.cm.get_cmap('tab20')((color % 20 + 0.05) / 20)

            self._draw_skeleton(ax, x, y, v, skeleton=skeleton, color=color)
            if self.show_box:
                score = scores[i] if scores is not None else None
                self._draw_box(ax, x, y, v, color, score)

            if texts is not None:
                self._draw_text(ax, x, y, v, texts[i], color)

    @staticmethod
    def _draw_box(ax, x, y, w, h, color, score=None, linewidth=1):
        if w < 5.0:
            x -= 2.0
            w += 4.0
        if h < 5.0:
            y -= 2.0
            h += 4.0
        ax.add_patch(
            matplotlib.patches.Rectangle(
                (x, y), w, h, fill=False, color=color, linewidth=linewidth))

        if score:
            ax.text(x, y - linewidth, '{:.4f}'.format(score), fontsize=8, color=color)

    @classmethod
    def _draw_text(cls, ax, x, y, v, text, color, *, subtext=None, alpha=1.0):
        if cls.font_size == 0:
            return
        if not np.any(v > 0):
            return

        coord_i = np.argsort(y[v > 0])
        if np.sum(v > 0) >= 2 \
           and y[v > 0][coord_i[1]] < y[v > 0][coord_i[0]] + 10:
            # second coordinate within 10 pixels
            f0 = 0.5 + 0.5 * (y[v > 0][coord_i[1]] - y[v > 0][coord_i[0]]) / 10.0
            coord_y = f0 * y[v > 0][coord_i[0]] + (1.0 - f0) * y[v > 0][coord_i[1]]
            coord_x = f0 * x[v > 0][coord_i[0]] + (1.0 - f0) * x[v > 0][coord_i[1]]
        else:
            coord_y = y[v > 0][coord_i[0]]
            coord_x = x[v > 0][coord_i[0]]

        bbox_config = {'facecolor': color, 'alpha': alpha * cls.textbox_alpha, 'linewidth': 0}
        ax.annotate(
            text,
            (coord_x, coord_y),
            fontsize=cls.font_size,
            xytext=(5.0, 5.0),
            textcoords='offset points',
            color=cls.text_color,
            bbox=bbox_config,
            alpha=alpha,
        )
        if subtext is not None:
            ax.annotate(
                subtext,
                (coord_x, coord_y),
                fontsize=cls.font_size * 5 // 8,
                xytext=(5.0, 18.0 + 3.0),
                textcoords='offset points',
                color=cls.text_color,
                bbox=bbox_config,
                alpha=alpha,
            )

    @staticmethod
    def _draw_scales(ax, xs, ys, vs, color, scales, alpha=1.0):
        for x, y, v, scale in zip(xs, ys, vs, scales):
            if v == 0.0:
                continue
            ax.add_patch(
                matplotlib.patches.Rectangle(
                    (x - scale / 2, y - scale / 2), scale, scale,
                    fill=False, color=color, alpha=alpha))

    @classmethod
    def _draw_joint_confidences(cls, ax, xs, ys, vs, color):
        for x, y, v in zip(xs, ys, vs):
            if v == 0.0:
                continue
            ax.annotate(
                '{:.0%}'.format(v),
                (x, y),
                fontsize=6,
                xytext=(0.0, 0.0),
                textcoords='offset points',
                verticalalignment='top',
                color=cls.text_color,
                bbox={'facecolor': color, 'alpha': 0.2, 'linewidth': 0, 'pad': 0.0},
            )

    def annotation(self, ax, ann, *, color=None, text=None, subtext=None, alpha=1.0):
        if color is None:
            color = 0
        if isinstance(color, (int, np.integer)):
            color = matplotlib.cm.get_cmap('tab20')((color % 20 + 0.05) / 20)

        text_is_score = False
        if text is None and hasattr(ann, 'id_'):
            text = '{}'.format(ann.id_)
        if text is None and getattr(ann, 'score', None):
            text = '{:.0%}'.format(ann.score)
            text_is_score = True
        if subtext is None and not text_is_score and getattr(ann, 'score', None):
            subtext = '{:.0%}'.format(ann.score)

        kps = ann.data
        assert kps.shape[1] == 3
        x = kps[:, 0] * self.xy_scale
        y = kps[:, 1] * self.xy_scale
        v = kps[:, 2]

        if self.show_frontier_order:
            frontier = set((s, e) for s, e in ann.frontier_order)
            frontier_skeleton_mask = [
                (s - 1, e - 1) in frontier or (e - 1, s - 1) in frontier
                for s, e in ann.skeleton
            ]
            frontier_skeleton = [se for se, m in zip(ann.skeleton, frontier_skeleton_mask) if m]
            self._draw_skeleton(ax, x, y, v, color='black', skeleton=frontier_skeleton,
                                linestyle='dotted', linewidth=1)

        skeleton_mask = None
        if self.show_only_decoded_connections:
            decoded_connections = set((jsi, jti) for jsi, jti, _, __ in ann.decoding_order)
            skeleton_mask = [
                (s - 1, e - 1) in decoded_connections or (e - 1, s - 1) in decoded_connections
                for s, e in ann.skeleton
            ]

        self._draw_skeleton(ax, x, y, v, color=color,
                            skeleton=ann.skeleton, skeleton_mask=skeleton_mask, alpha=alpha)

        if self.show_joint_scales and ann.joint_scales is not None:
            self._draw_scales(ax, x, y, v, color, ann.joint_scales, alpha=alpha)

        if self.show_joint_confidences:
            self._draw_joint_confidences(ax, x, y, v, color)

        if self.show_box:
            x_, y_, w_, h_ = [v * self.xy_scale for v in ann.bbox()]
            self._draw_box(ax, x_, y_, w_, h_, color, ann.score)

        if text is not None:
            self._draw_text(ax, x, y, v, text, color, subtext=subtext, alpha=alpha)

        if self.show_decoding_order and hasattr(ann, 'decoding_order'):
            self._draw_decoding_order(ax, ann.decoding_order)

    @staticmethod
    def _draw_decoding_order(ax, decoding_order):
        for step_i, (jsi, jti, jsxyv, jtxyv) in enumerate(decoding_order):
            ax.plot([jsxyv[0], jtxyv[0]], [jsxyv[1], jtxyv[1]], '--', color='black')
            ax.text(0.5 * (jsxyv[0] + jtxyv[0]), 0.5 * (jsxyv[1] + jtxyv[1]),
                    '{}: {} -> {}'.format(step_i, jsi, jti), fontsize=8,
                    color='white', bbox={'facecolor': 'black', 'alpha': 0.5, 'linewidth': 0})



import importlib
import os
import torch


def register_ops():
    lib_dir = os.path.dirname(__file__)
    if hasattr(os, 'add_dll_directory'):  # for Windows
        import ctypes  # pylint: disable=import-outside-toplevel

        kernel32 = ctypes.WinDLL('kernel32.dll', use_last_error=True)
        if hasattr(kernel32, 'AddDllDirectory'):
            kernel32.AddDllDirectory.restype = ctypes.c_void_p

        os.add_dll_directory(lib_dir)  # pylint: disable=no-member

    loader_details = (
        importlib.machinery.ExtensionFileLoader,
        importlib.machinery.EXTENSION_SUFFIXES
    )

    extfinder = importlib.machinery.FileFinder(lib_dir, loader_details)
    ext_specs = extfinder.find_spec("_cpp")
    if ext_specs is None:
        raise ImportError
    torch.ops.load_library(ext_specs.origin)



import copy
import math
import logging

import torch
import torchvision

from .preprocess import Preprocess

LOG = logging.getLogger(__name__)


class CenterPad(Preprocess):
    """Pad to a square of given size."""

    def __init__(self, target_size: int):
        if isinstance(target_size, int):
            target_size = (target_size, target_size)
        self.target_size = target_size

    def __call__(self, image, anns, meta):
        meta = copy.deepcopy(meta)
        anns = copy.deepcopy(anns)

        LOG.debug('valid area before pad: %s, image size = %s, target size = %s',
                  meta['valid_area'], image.size, self.target_size)
        image, anns, ltrb = self.center_pad(image, anns)
        meta['offset'] -= ltrb[:2]
        meta['valid_area'][:2] += ltrb[:2]
        LOG.debug('valid area after pad: %s, image size = %s', meta['valid_area'], image.size)

        return image, anns, meta

    def center_pad(self, image, anns):
        w, h = image.size

        left = int((self.target_size[0] - w) / 2.0)
        top = int((self.target_size[1] - h) / 2.0)
        left = max(0, left)
        top = max(0, top)

        right = self.target_size[0] - w - left
        bottom = self.target_size[1] - h - top
        right = max(0, right)
        bottom = max(0, bottom)
        ltrb = (left, top, right, bottom)
        LOG.debug('pad with %s', ltrb)

        # pad image
        fill_value = int(torch.randint(0, 255, (1,)).item())
        image = torchvision.transforms.functional.pad(
            image, ltrb, fill=(fill_value, fill_value, fill_value))

        # pad annotations
        for ann in anns:
            ann['keypoints'][:, 0] += ltrb[0]
            ann['keypoints'][:, 1] += ltrb[1]
            ann['bbox'][0] += ltrb[0]
            ann['bbox'][1] += ltrb[1]

        return image, anns, ltrb


class CenterPadTight(Preprocess):
    def __init__(self, multiple):
        self.multiple = multiple

    def __call__(self, image, anns, meta):
        meta = copy.deepcopy(meta)
        anns = copy.deepcopy(anns)

        LOG.debug('valid area before pad: %s, image size = %s', meta['valid_area'], image.size)
        image, anns, ltrb = self.center_pad(image, anns)
        meta['offset'] -= ltrb[:2]
        meta['valid_area'][:2] += ltrb[:2]
        LOG.debug('valid area after pad: %s, image size = %s', meta['valid_area'], image.size)

        return image, anns, meta

    def center_pad(self, image, anns):
        w, h = image.size
        target_width = math.ceil((w - 1) / self.multiple) * self.multiple + 1
        target_height = math.ceil((h - 1) / self.multiple) * self.multiple + 1

        left = int((target_width - w) / 2.0)
        top = int((target_height - h) / 2.0)
        left = max(0, left)
        top = max(0, top)

        right = target_width - w - left
        bottom = target_height - h - top
        right = max(0, right)
        bottom = max(0, bottom)
        ltrb = (left, top, right, bottom)
        LOG.debug('pad with %s', ltrb)

        # pad image
        image = torchvision.transforms.functional.pad(
            image, ltrb, fill=(124, 116, 104))

        # pad annotations
        for ann in anns:
            ann['keypoints'][:, 0] += ltrb[0]
            ann['keypoints'][:, 1] += ltrb[1]
            ann['bbox'][0] += ltrb[0]
            ann['bbox'][1] += ltrb[1]

        return image, anns, ltrb


class SquarePad(Preprocess):
    def __call__(self, image, anns, meta):
        center_pad = CenterPad(max(image.size))
        return center_pad(image, anns, meta)



import argparse

import torch

import openpifpaf

from . import datasets, metric
from .posetrack2018 import Posetrack2018

from .constants import (
    KEYPOINTS,
    SIGMAS,
    UPRIGHT_POSE,
    SKELETON,
    DENSER_CONNECTIONS,
)


class Posetrack2017(openpifpaf.datasets.DataModule):
    # cli configurable
    train_annotations = 'data-posetrack2017/annotations/train/*.json'
    val_annotations = 'data-posetrack2017/annotations/val/*.json'
    eval_annotations = val_annotations
    data_root = 'data-posetrack2017'

    def __init__(self):
        super().__init__()

        cif = openpifpaf.headmeta.TSingleImageCif(
            'cif', 'posetrack2018',
            keypoints=KEYPOINTS,
            sigmas=SIGMAS,
            pose=UPRIGHT_POSE,
            draw_skeleton=SKELETON,
        )
        caf = openpifpaf.headmeta.TSingleImageCaf(
            'caf', 'posetrack2018',
            keypoints=KEYPOINTS,
            sigmas=SIGMAS,
            pose=UPRIGHT_POSE,
            skeleton=SKELETON,
        )
        dcaf = openpifpaf.headmeta.TSingleImageCaf(
            'dcaf', 'posetrack2018',
            keypoints=KEYPOINTS,
            sigmas=SIGMAS,
            pose=UPRIGHT_POSE,
            skeleton=DENSER_CONNECTIONS,
            sparse_skeleton=SKELETON,
            only_in_field_of_view=True,
        )
        tcaf = openpifpaf.headmeta.Tcaf(
            'tcaf', 'posetrack2018',
            keypoints=KEYPOINTS,
            sigmas=SIGMAS,
            pose=UPRIGHT_POSE,
            draw_skeleton=SKELETON,
            only_in_field_of_view=True,
        )

        cif.upsample_stride = Posetrack2018.upsample_stride
        caf.upsample_stride = Posetrack2018.upsample_stride
        dcaf.upsample_stride = Posetrack2018.upsample_stride
        tcaf.upsample_stride = Posetrack2018.upsample_stride
        self.head_metas = [cif, caf, dcaf, tcaf]

        if Posetrack2018.ablation_without_tcaf:
            self.head_metas = [cif, caf, dcaf]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module Posetrack2017')

        group.add_argument('--posetrack2017-eval-annotations',
                           default=cls.eval_annotations,
                           help='eval annotations')
        group.add_argument('--posetrack2017-data-root',
                           default=cls.data_root,
                           help='data root')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # posetrack2017 specific
        cls.eval_annotations = args.posetrack2017_eval_annotations
        cls.data_root = args.posetrack2017_data_root

    def train_loader(self):
        raise NotImplementedError

    def val_loader(self):
        raise NotImplementedError

    def _eval_preprocess(self):
        return openpifpaf.transforms.Compose([
            *Posetrack2018.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToKpAnnotations(
                    ['person'],
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                openpifpaf.transforms.ToCrowdAnnotations(['person']),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = datasets.Posetrack2017(
            annotation_files=self.eval_annotations,
            data_root=self.data_root,
            preprocess=self._eval_preprocess(),
        )
        eval_loader = torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)
        return openpifpaf.datasets.LoaderWithReset(eval_loader, 'annotation_file')

    def metrics(self):
        eval_data = datasets.Posetrack2017(
            annotation_files=self.eval_annotations,
            data_root=self.data_root,
            preprocess=self._eval_preprocess(),
        )
        return [metric.Posetrack(
            images=eval_data.meta_images(),
            categories=eval_data.meta_categories(),
            ground_truth=self.eval_annotations,
            output_format='2017',
        )]



import argparse
import logging

import PIL
import torch

from . import datasets, decoder, network, transforms, visualizer

LOG = logging.getLogger(__name__)


class Predictor:
    """Convenience class to predict from various inputs with a common configuration."""

    batch_size = 1  #: batch size
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')  #: device
    fast_rescaling = True  #: fast rescaling
    loader_workers = None  #: loader workers
    long_edge = None  #: long edge

    def __init__(self, checkpoint=None, head_metas=None, *,
                 json_data=False,
                 visualize_image=False,
                 visualize_processed_image=False):
        if checkpoint is not None:
            network.Factory.checkpoint = checkpoint
        self.json_data = json_data
        self.visualize_image = visualize_image
        self.visualize_processed_image = visualize_processed_image

        self.model_cpu, _ = network.Factory().factory(head_metas=head_metas)
        self.model = self.model_cpu.to(self.device)
        if self.device.type == 'cuda' and torch.cuda.device_count() > 1:
            LOG.info('Using multiple GPUs: %d', torch.cuda.device_count())
            self.model = torch.nn.DataParallel(self.model)
            self.model.base_net = self.model_cpu.base_net
            self.model.head_nets = self.model_cpu.head_nets

        self.preprocess = self._preprocess_factory()
        self.processor = decoder.factory(self.model_cpu.head_metas)

        self.last_decoder_time = 0.0
        self.last_nn_time = 0.0
        self.total_nn_time = 0.0
        self.total_decoder_time = 0.0
        self.total_images = 0

        LOG.info('neural network device: %s (CUDA available: %s, count: %d)',
                 self.device, torch.cuda.is_available(), torch.cuda.device_count())

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser, *,
            skip_batch_size=False, skip_loader_workers=False):
        """Add command line arguments.

        When using this class together with datasets (e.g. in eval),
        skip the cli arguments for batch size and loader workers as those
        will be provided via the datasets module.
        """
        group = parser.add_argument_group('Predictor')

        if not skip_batch_size:
            group.add_argument('--batch-size', default=cls.batch_size, type=int,
                               help='processing batch size')

        if not skip_loader_workers:
            group.add_argument('--loader-workers', default=cls.loader_workers, type=int,
                               help='number of workers for data loading')

        group.add_argument('--long-edge', default=cls.long_edge, type=int,
                           help='rescale the long side of the image (aspect ratio maintained)')
        group.add_argument('--precise-rescaling', dest='fast_rescaling',
                           default=True, action='store_false',
                           help='use more exact image rescaling (requires scipy)')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Configure from command line parser."""
        cls.batch_size = args.batch_size
        cls.device = args.device
        cls.fast_rescaling = args.fast_rescaling
        cls.loader_workers = args.loader_workers
        cls.long_edge = args.long_edge

    def _preprocess_factory(self):
        rescale_t = None
        if self.long_edge:
            rescale_t = transforms.RescaleAbsolute(self.long_edge, fast=self.fast_rescaling)

        pad_t = None
        if self.batch_size > 1:
            assert self.long_edge, '--long-edge must be provided for batch size > 1'
            pad_t = transforms.CenterPad(self.long_edge)
        else:
            pad_t = transforms.CenterPadTight(16)

        return transforms.Compose([
            transforms.NormalizeAnnotations(),
            rescale_t,
            pad_t,
            transforms.EVAL_TRANSFORM,
        ])

    def dataset(self, data):
        """Predict from a dataset."""
        loader_workers = self.loader_workers
        if loader_workers is None:
            loader_workers = self.batch_size if len(data) > 1 else 0

        dataloader = torch.utils.data.DataLoader(
            data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.device.type != 'cpu',
            num_workers=loader_workers,
            collate_fn=datasets.collate_images_anns_meta)

        yield from self.dataloader(dataloader)

    def enumerated_dataloader(self, enumerated_dataloader):
        """Predict from an enumerated dataloader."""
        for batch_i, item in enumerated_dataloader:
            if len(item) == 3:
                processed_image_batch, gt_anns_batch, meta_batch = item
                image_batch = [None for _ in processed_image_batch]
            elif len(item) == 4:
                image_batch, processed_image_batch, gt_anns_batch, meta_batch = item
            if self.visualize_processed_image:
                visualizer.Base.processed_image(processed_image_batch[0])

            pred_batch = self.processor.batch(self.model, processed_image_batch, device=self.device)
            self.last_decoder_time = self.processor.last_decoder_time
            self.last_nn_time = self.processor.last_nn_time
            self.total_decoder_time += self.processor.last_decoder_time
            self.total_nn_time += self.processor.last_nn_time
            self.total_images += len(processed_image_batch)

            # un-batch
            for image, pred, gt_anns, meta in \
                    zip(image_batch, pred_batch, gt_anns_batch, meta_batch):
                LOG.info('batch %d: %s', batch_i, meta.get('file_name', 'no-file-name'))

                # load the original image if necessary
                if self.visualize_image:
                    visualizer.Base.image(image, meta=meta)

                pred = [ann.inverse_transform(meta) for ann in pred]
                gt_anns = [ann.inverse_transform(meta) for ann in gt_anns]

                if self.json_data:
                    pred = [ann.json_data() for ann in pred]

                yield pred, gt_anns, meta

    def dataloader(self, dataloader):
        """Predict from a dataloader."""
        yield from self.enumerated_dataloader(enumerate(dataloader))

    def image(self, file_name):
        """Predict from an image file name."""
        return next(iter(self.images([file_name])))

    def images(self, file_names, **kwargs):
        """Predict from image file names."""
        data = datasets.ImageList(
            file_names, preprocess=self.preprocess, with_raw_image=True)
        yield from self.dataset(data, **kwargs)

    def pil_image(self, image):
        """Predict from a Pillow image."""
        return next(iter(self.pil_images([image])))

    def pil_images(self, pil_images, **kwargs):
        """Predict from Pillow images."""
        data = datasets.PilImageList(
            pil_images, preprocess=self.preprocess, with_raw_image=True)
        yield from self.dataset(data, **kwargs)

    def numpy_image(self, image):
        """Predict from a numpy image."""
        return next(iter(self.numpy_images([image])))

    def numpy_images(self, numpy_images, **kwargs):
        """Predict from numpy images."""
        data = datasets.NumpyImageList(
            numpy_images, preprocess=self.preprocess, with_raw_image=True)
        yield from self.dataset(data, **kwargs)

    def image_file(self, file_pointer):
        """Predict from an opened image file pointer."""
        pil_image = PIL.Image.open(file_pointer).convert('RGB')
        return self.pil_image(pil_image)



from collections import defaultdict
import logging

import torch

LOG = logging.getLogger(__name__)


class RunningCache(torch.nn.Module):
    def __init__(self, cached_items):
        super().__init__()

        self.cached_items = cached_items
        self.duration = abs(min(cached_items)) + 1
        self.cache = [None for _ in range(self.duration)]
        self.index = 0

        LOG.debug('running cache of length %d', len(self.cache))

    def incr(self):
        self.index = (self.index + 1) % self.duration

    def get_index(self, index):
        while index < 0:
            index += self.duration
        while index >= self.duration:
            index -= self.duration
        LOG.debug('retrieving cache at index %d', index)

        v = self.cache[index]
        if v is not None:
            v = v.detach()
        return v

    def get(self):
        return [self.get_index(i + self.index) for i in self.cached_items]

    def set_next(self, data):
        self.incr()
        self.cache[self.index] = data
        LOG.debug('set new data at index %d', self.index)
        return self

    def forward(self, *args):
        LOG.debug('----------- running cache --------------')
        x = args[0]

        o = []
        for x_i in x:
            o += self.set_next(x_i).get()

        if any(oo is None for oo in o):
            o = [oo if oo is not None else o[0] for oo in o]

        # drop images of the wrong size (determine size by majority vote)
        if len(o) >= 2:
            image_sizes = [tuple(oo.shape[-2:]) for oo in o]
            if not all(ims == image_sizes[0] for ims in image_sizes[1:]):
                freq = defaultdict(int)
                for ims in image_sizes:
                    freq[ims] += 1
                max_freq = max(freq.values())
                ref_image_size = next(iter(ims for ims, f in freq.items() if f == max_freq))

                for i, ims in enumerate(image_sizes):
                    if ims == ref_image_size:
                        continue
                    for s in range(1, len(image_sizes)):
                        target_i = (i + s) % len(image_sizes)
                        if image_sizes[target_i] == ref_image_size:
                            break
                    LOG.warning('replacing %d (%s) with %d (%s) for ref %s',
                                i, ims,
                                target_i, image_sizes[target_i],
                                ref_image_size)
                    o[i] = o[target_i]

        return torch.stack(o)



"""Configuring and visualizing log files."""

import argparse
from collections import defaultdict
import datetime
import json
import logging
from pprint import pprint

import numpy as np
import pysparkling

from . import logger, metric, show, __version__

try:
    import matplotlib
except ImportError:
    matplotlib = None

LOG = logging.getLogger(__name__)


def optionally_shaded(ax, x, y, *, color, label, **kwargs):
    epochs = x[-1] - x[0] if len(x) > 1 else 0.0
    stride = int(len(x) / epochs / 30.0) if len(x) > 30 else 1  # 30 per epoch
    if stride > 1:
        x_binned = np.array([x[i] for i in range(0, len(x), stride)][:-1])
        y_binned = np.stack([y[i:i + stride] for i in range(0, len(x), stride)][:-1])
        y_mean = np.mean(y_binned, axis=1)
        y_min = np.min(y_binned, axis=1)
        y_max = np.max(y_binned, axis=1)
        ax.plot(x_binned, y_mean, color=color, label=label, **kwargs)
        ax.fill_between(x_binned, y_min, y_max, alpha=0.2, facecolor=color)
    else:
        LOG.debug('not shading: entries = %d, epochs = %f', len(x), epochs)
        ax.plot(x, y, color=color, label=label, **kwargs)


def fractional_epoch(row, *, default=None):
    """Given a data row, compute the fractional epoch taking batch into account.

    Example:
        Epoch 1 at batch 30 out of 100 batches per epoch would return
        epoch 1.3.
    """

    if 'epoch' not in row:
        return default
    if 'batch' not in row:
        return row.get('epoch')
    return row.get('epoch') + row.get('batch') / row.get('n_batches')


class Plots():
    def __init__(self, log_files, labels=None, *,
                 output_prefix=None, first_epoch=0.0, share_y=True):
        self.log_files = log_files
        self.labels = labels or [lf.replace('outputs/', '') for lf in log_files]
        self.output_prefix = output_prefix or log_files[-1] + '.'
        self.first_epoch = first_epoch
        self.share_y = share_y

        self.datas = [self.read_log(f) for f in log_files]

    def read_log(self, path):
        LOG.debug('reading file %s', path)
        sc = pysparkling.Context()
        return (sc
                .textFile(path)
                .filter(lambda line: line.startswith(('{', 'json:')) and line.endswith('}'))
                .map(lambda line: json.loads(line.strip('json:')))
                .filter(lambda data: fractional_epoch(data, default=np.inf) >= self.first_epoch)
                .groupBy(lambda data: data.get('type'))
                .collectAsMap())

    def process(self):
        return {label: data['process']
                for data, label in zip(self.datas, self.labels)}

    def field_names(self):
        def migrate(field_name):
            """to process older (pre v0.12) log files"""
            if field_name.startswith(('cif.', 'caf.', 'caf25.')):
                return 'cocokp.{}'.format(field_name)
            if field_name.startswith('cifdet.'):
                return 'cocodet.{}'.format(field_name)
            return field_name
        return {label: [migrate(f) for f in data['config'][0]['field_names']]
                for data, label in zip(self.datas, self.labels)}

    def process_arguments(self):
        if not self.datas[0]:
            raise Exception('no data')
        return {label: data['process'][0]['argv'][1:]
                for data, label in zip(self.datas, self.labels)}

    def time(self, ax):
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)

            if 'train' in data:
                x = np.array([fractional_epoch(row) for row in data['train']])
                y = [datetime.datetime.strptime(row.get('asctime')[:-4], '%Y-%m-%d %H:%M:%S')
                     for row in data['train']]
                y = [(yi - y[0]).total_seconds() / 3600.0 for yi in y]
                ax.plot(x, y, color=color, label=label)

        ax.set_xlabel('epoch')
        ax.set_ylabel('time [h]')
        ax.legend(loc='upper left')

    def epoch_time(self, ax):
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)

            y0 = None
            if 'train' in data:
                row = data['train'][0]
                y0 = datetime.datetime.strptime(row.get('asctime')[:-4], '%Y-%m-%d %H:%M:%S')

            if 'train-epoch' in data:
                x = [row.get('epoch') for row in data['train-epoch']]
                y = [datetime.datetime.strptime(row.get('asctime')[:-4], '%Y-%m-%d %H:%M:%S')
                     for row in data['train-epoch']]
                if y0 is not None:
                    x = [0] + x
                    y = [y0] + y
                y = [(yi - prev_yi).total_seconds() / 60.0
                     for prev_yi, yi in zip(y[:-1], y[1:])]
                ax.plot(x[1:], y, color=color, label=label)

        ax.set_xlabel('epoch')
        ax.set_ylabel('epoch-time [min]')
        ax.legend(loc='lower right')

    def lr(self, ax):
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)

            if 'train' in data:
                x = [fractional_epoch(row) for row in data['train']]
                y = [row.get('lr') for row in data['train']]
                ax.plot(x, y, color=color, label=label)

        ax.set_xlabel('epoch')
        ax.set_ylabel('learning rate')
        ax.set_yscale('log', nonpositive='clip')
        ax.legend(loc='upper left')

    def epoch_loss(self, ax):
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)

            if 'val-epoch' in data:
                x = np.array([row.get('epoch') for row in data['val-epoch']])
                y = np.array([row.get('loss') for row in data['val-epoch']], dtype=np.float64)
                ax.plot(x, y, 'o-', color=color, markersize=2, label=label)

            if 'train-epoch' in data:
                x = np.array([row.get('epoch') for row in data['train-epoch']])
                y = np.array([row.get('loss') for row in data['train-epoch']], dtype=np.float64)
                m = x > 0
                ax.plot(x[m], y[m], 'x:', color=color, markersize=2)

        ax.set_xlabel('epoch')
        ax.set_ylabel('loss')
        # ax.set_ylim(0.0, 4.0)
        # if min(y) > -0.1:
        #     ax.set_yscale('log', nonpositive='clip')
        ax.grid(linestyle='dotted')
        ax.legend(loc='upper right')
        ax.text(0.01, 1.01, 'train (cross-dotted), val (dot-solid)',
                transform=ax.transAxes, size='x-small')

    def epoch_head(self, ax, field_name):
        field_names = self.field_names()
        last_five_y = []
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)
            if field_name not in field_names[label]:
                continue
            field_i = field_names[label].index(field_name)
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)

            if 'val-epoch' in data:
                x = np.array([row.get('epoch') for row in data['val-epoch']])
                y = np.array([row.get('head_losses')[field_i]
                              for row in data['val-epoch']], dtype=np.float64)
                m = np.logical_not(np.isnan(y))
                ax.plot(x[m], y[m], 'o-', color=color, markersize=2, label=label)
                last_five_y.append(y[m][-5:])

            if 'train-epoch' in data:
                x = np.array([row.get('epoch') for row in data['train-epoch']])
                y = np.array([row.get('head_losses')[field_i]
                              for row in data['train-epoch']], dtype=np.float64)
                m = np.logical_not(np.isnan(y))
                ax.plot(x[m], y[m], 'x:', color=color, markersize=2)
                last_five_y.append(y[m][-5:])

        if not last_five_y:
            return
        ax.set_xlabel('epoch')
        ax.set_ylabel(field_name, fontsize=8 if len(field_name) < 30 else 5)
        last_five_y = np.concatenate(last_five_y)
        if not self.share_y and last_five_y.size >= 2:
            ax.set_ylim(np.min(last_five_y), np.max(last_five_y))
        # ax.set_ylim(0.0, 1.0)
        # if min(y) > -0.1:
        #     ax.set_yscale('log', nonpositive='clip')
        ax.grid(linestyle='dotted')
        # ax.legend(loc='upper right')
        ax.text(0.01, 1.01, 'train (cross-dotted), val (dot-solid)',
                transform=ax.transAxes, size='x-small')

    def preprocess_time(self, ax):
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)

            if 'train' in data:
                # skip batch 0 as it has corrupted data_time
                x = np.array([fractional_epoch(row)
                              for row in data['train']
                              if row.get('batch', 1) > 0])
                y = np.array([row.get('data_time') / row.get('time') * 100.0
                              for row in data['train']
                              if row.get('batch', 1) > 0], dtype=np.float64)
                optionally_shaded(ax, x, y, color=color, label=label)

        ax.set_xlabel('epoch')
        ax.set_ylabel('data preprocessing time [%]')
        ax.set_ylim(0, 100)
        ax.legend(loc='upper right')

    def train(self, ax):
        miny = 0.0
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)
            if 'train' in data:
                xy_all = defaultdict(list)
                for row in data['train']:
                    xy_all[row.get('loss_index', 0)].append(
                        (fractional_epoch(row), row.get('loss'))
                    )
                for loss_index, xy in xy_all.items():
                    x = np.array([x for x, _ in xy])
                    y = np.array([y for _, y in xy], dtype=np.float64)
                    miny = min(miny, np.min(y))

                    kwargs = {}
                    this_label = label
                    if loss_index != 0:
                        kwargs['alpha'] = 0.5
                        this_label = '{} ({})'.format(label, loss_index)
                    optionally_shaded(ax, x, y, color=color, label=this_label, **kwargs)

        ax.set_xlabel('epoch')
        ax.set_ylabel('training loss')
        # ax.set_ylim(0, 8)
        if miny > -0.1:
            ax.set_yscale('log', nonpositive='clip')
        ax.grid(linestyle='dotted')
        ax.legend(loc='upper right')

    def train_head(self, ax, field_name):
        field_names = self.field_names()
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)
            if field_name not in field_names[label]:
                continue
            field_i = field_names[label].index(field_name)

            if 'train' in data:
                x = np.array([fractional_epoch(row) for row in data['train']])
                y = np.array([row.get('head_losses')[field_i]
                              for row in data['train']], dtype=np.float64)
                m = np.logical_not(np.isnan(y))
                optionally_shaded(ax, x[m], y[m], color=color, label=label)

        ax.set_xlabel('epoch')
        ax.set_ylabel(field_name, fontsize=8 if len(field_name) < 30 else 5)
        # ax.set_ylim(3e-3, 3.0)
        if not self.share_y and min(y) > -0.1:
            ax.set_yscale('log', nonpositive='clip')
        ax.grid(linestyle='dotted')
        # ax.legend(loc='upper right')

    def mtl_sigma(self, ax, field_name):
        field_names = self.field_names()
        for color_i, (data, label) in enumerate(zip(self.datas, self.labels)):
            color = matplotlib.cm.get_cmap('tab10')((color_i % 10 + 0.05) / 10)
            if field_name not in field_names[label]:
                continue
            field_i = field_names[label].index(field_name)

            if 'train' in data:
                x = np.array([fractional_epoch(row) for row in data['train']])
                y = np.array([row['mtl_sigmas'][field_i] if 'mtl_sigmas' in row else np.nan
                              for row in data['train']], dtype=np.float64)
                m = np.logical_not(np.isnan(y))
                optionally_shaded(ax, x[m], y[m], color=color, label=label)

        ax.set_xlabel('epoch')
        ax.set_ylabel(field_name)
        ax.set_ylim(-0.1, 1.1)
        if min(y) > -0.1:
            ax.set_ylim(3e-3, 3.0)
            ax.set_yscale('log', nonpositive='clip')
        ax.grid(linestyle='dotted')
        # ax.legend(loc='upper right')

    def print_last_line(self):
        for data, label in zip(self.datas, self.labels):
            if 'train' in data:
                print('{}: {}'.format(label, data['train'][-1]))

    def show_all(self, show_mtl_sigmas=False):
        pprint(self.process_arguments())

        all_field_names = [f for fs in self.field_names().values() for f in fs]
        rows = defaultdict(list)
        for f in all_field_names:
            dataset_name, head_name = f.split('.')[:2]
            row_name = dataset_name + '.' + head_name
            if f not in rows[row_name]:
                rows[row_name].append(f)
        n_rows = len(rows)
        n_cols = max(len(r) for r in rows.values())
        multi_figsize = (5 * n_cols, 2.5 * n_rows)

        with show.canvas() as ax:
            self.time(ax)

        with show.canvas() as ax:
            self.epoch_time(ax)

        with show.canvas() as ax:
            self.lr(ax)

        with show.canvas(nrows=n_rows, ncols=n_cols, squeeze=False,
                         figsize=multi_figsize,
                         sharey=self.share_y, sharex=True) as axs:
            for row_i, row in enumerate(rows.values()):
                for col_i, field_name in enumerate(row):
                    self.epoch_head(axs[row_i, col_i], field_name)

        with show.canvas() as ax:
            self.epoch_loss(ax)

        with show.canvas() as ax:
            self.preprocess_time(ax)

        with show.canvas(nrows=n_rows, ncols=n_cols, squeeze=False,
                         figsize=multi_figsize,
                         sharey=self.share_y, sharex=True) as axs:
            for row_i, row in enumerate(rows.values()):
                for col_i, field_name in enumerate(row):
                    self.train_head(axs[row_i, col_i], field_name)

        if show_mtl_sigmas:
            with show.canvas(nrows=n_rows, ncols=n_cols, squeeze=False,
                             figsize=multi_figsize,
                             sharey=self.share_y, sharex=True) as axs:
                for row_i, row in enumerate(rows.values()):
                    for col_i, field_name in enumerate(row):
                        self.mtl_sigma(axs[row_i, col_i], field_name)

        with show.canvas() as ax:
            self.train(ax)

        self.print_last_line()


class EvalPlots():
    #: Eval files come with text labels. This is a translation to prettier labels
    #: for matplotlib axes.
    text_to_latex_labels = {
        'AP0.5': 'AP$^{0.50}$',
        'AP0.75': 'AP$^{0.75}$',
        'APS': 'AP$^{S}$',
        'APM': 'AP$^{M}$',
        'APL': 'AP$^{L}$',
        'ART1': 'AR@1',
        'ART10': 'AR@10',
        'AR0.5': 'AR$^{0.50}$',
        'AR0.75': 'AR$^{0.75}$',
        'ARS': 'AR$^{S}$',
        'ARM': 'AR$^{M}$',
        'ARL': 'AR$^{L}$',
    }

    def __init__(self, log_files, file_suffix, *,
                 labels=None, output_prefix=None,
                 decoder=0, legend_last_ap=True,
                 first_epoch=0.0, share_y=True):
        self.file_suffix = file_suffix
        self.decoder = decoder
        self.legend_last_ap = legend_last_ap
        self.first_epoch = first_epoch
        self.share_y = share_y

        self.datas = [self.read_log(f) for f in log_files]
        self.labels = labels or [lf.replace('outputs/', '') for lf in log_files]
        self.output_prefix = output_prefix or log_files[-1] + '.'

    def read_log(self, path):
        sc = pysparkling.Context()

        # modify individual file names and comma-seperated filenames
        files = path.split(',')
        files = ','.join([
            '{}.epoch???{}'.format(f[:-4], self.file_suffix)
            for f in files
        ])
        LOG.debug('reading files %s', files)

        def epoch_from_filename(filename):
            i = filename.find('epoch')
            return int(filename[i + 5:i + 8])

        def migrate(data):
            # earlier versions did not contain 'dataset'
            if 'dataset' not in data and len(data['stats']) == 10:
                data['dataset'] = 'cocokp'
            if 'dataset' not in data and len(data['stats']) == 12:
                data['dataset'] = 'cocodet'

            # earlier versions did not contain 'text_labels'
            if 'text_labels' not in data and len(data['stats']) == 10:
                data['text_labels'] = metric.Coco.text_labels_keypoints
            if 'text_labels' not in data and len(data['stats']) == 12:
                data['text_labels'] = metric.Coco.text_labels_bbox

            return data

        return (sc
                .wholeTextFiles(files)
                .map(lambda k_c: (
                    epoch_from_filename(k_c[0]),
                    json.loads(k_c[1]),
                ))
                .filter(lambda k_c: k_c[0] >= self.first_epoch and k_c[1]['stats'])
                .mapValues(migrate)
                .sortByKey()
                .collect())

    def metrics(self):
        all_metrics_by_datasets = defaultdict(list)
        for data in self.datas:
            if not data:
                continue
            dataset = data[0][1]['dataset']
            for m in data[0][1]['text_labels']:
                if m in all_metrics_by_datasets[dataset]:
                    continue
                all_metrics_by_datasets[dataset].append(m)
        return all_metrics_by_datasets

    def fill_metric(self, ax, dataset, metric_name):
        for data, label in zip(self.datas, self.labels):
            if not data:
                continue
            if data[0][1]['dataset'] != dataset:
                continue
            if metric_name not in data[0][1]['text_labels']:
                continue

            entry = data[0][1]['text_labels'].index(metric_name)
            LOG.debug('processing %s at entry %d', metric_name, entry)
            if self.legend_last_ap:
                last_main_value = data[-1][1]['stats'][0]
                main_name = data[0][1]['text_labels'][0]
                main_label = self.text_to_latex_labels.get(main_name, main_name)
                label = '{} ({}={:.1%})'.format(label, main_label, last_main_value)
            x = np.array([e for e, _ in data])
            y = np.array([d['stats'][entry] if entry < len(d['stats']) else np.nan
                          for _, d in data])
            ax.plot(x, y, 'o-', label=label, markersize=2)

        ax.set_xlabel('epoch')
        ax.set_ylabel('{} {}'.format(
            dataset,
            self.text_to_latex_labels.get(metric_name, metric_name),
        ))
        ax.grid(linestyle='dotted')
        # ax.legend(loc='upper right')

    def frame_ops(self, ax, entry):
        assert entry in (0, 1)

        s = 1e9 if entry == 0 else 1e6
        for data, label in zip(self.datas, self.labels):
            if not data:
                continue
            x = np.array([d.get('count_ops', [0, 0])[entry] / s for _, d in data])[-1]
            if x == 0.0:
                continue
            y = np.array([d['stats'][0] for _, d in data])[-1]
            ax.plot([x], [y], 'o', label=label, markersize=10)
            ax.annotate(
                label if len(label) < 20 else label.split('-')[0],
                (x, y),
                xytext=(0.0, -5.0),
                textcoords='offset points',
                rotation=90,
                horizontalalignment='center', verticalalignment='top',
            )

        # ax.set_ylim(bottom=0.56)
        ax.set_xlabel('GMACs' if entry == 0 else 'million parameters')
        ax.set_ylabel('AP')
        ax.grid(linestyle='dotted')
        # ax.legend(loc='lower right')

    def show_all(self):
        # layouting: a dataset can span one or two rows
        all_metrics = self.metrics()
        all_rows_nested = {
            dataset: (
                [metrics]
                if len(metrics) <= 6
                else [metrics[:-len(metrics) // 2],
                      metrics[-len(metrics) // 2:]]
            )
            for dataset, metrics in all_metrics.items()
        }
        all_rows = [
            [(dataset, metric) for metric in row]
            for dataset, rows in all_rows_nested.items()
            for row in rows
        ]
        if not all_rows:
            return
        nrows = len(all_rows)
        ncols = max(len(row) for row in all_rows)

        # plot
        with show.canvas(nrows=nrows, ncols=ncols, figsize=(4 * ncols, 3 * nrows),
                         sharex=True, sharey=self.share_y, squeeze=False) as axs:
            for ax_row, metric_row in zip(axs, all_rows):
                for ax, (dataset, metric_name) in zip(ax_row, metric_row):
                    self.fill_metric(ax, dataset, metric_name)
                ax_row[len(metric_row) - 1].legend(fontsize=5, loc='lower right')

        with show.canvas(nrows=1, ncols=2, figsize=(10, 5),
                         sharey=self.share_y) as axs:
            self.frame_ops(axs[0], 0)
            self.frame_ops(axs[1], 1)


def main():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.logs',
        usage='%(prog)s [options] log_files',
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=__version__))

    logger.cli(parser)
    show.cli(parser)

    parser.add_argument('log_file', nargs='+',
                        help='path to log file(s)')
    parser.add_argument('--label', nargs='+',
                        help='label(s) in the same order as files')
    parser.add_argument('--eval-suffix', default='.eval-*.stats.json',
                        help='suffix of evaluation files to look for')
    parser.add_argument('--first-epoch', default=1e-6, type=float,
                        help='epoch (can be float) of first data point to plot')
    parser.add_argument('--no-share-y', dest='share_y',
                        default=True, action='store_false',
                        help='dont share y access')
    parser.add_argument('-o', '--output', default=None,
                        help='output prefix (default is log_file + .)')
    parser.add_argument('--show-mtl-sigmas', default=False, action='store_true')
    args = parser.parse_args()

    logger.configure(args, LOG)
    show.configure(args)

    if args.output is None:
        args.output = args.log_file[-1] + '.'

    EvalPlots(args.log_file, args.eval_suffix,
              labels=args.label,
              output_prefix=args.output,
              first_epoch=args.first_epoch,
              share_y=args.share_y,
              ).show_all()
    Plots(args.log_file, args.label,
          output_prefix=args.output,
          first_epoch=args.first_epoch,
          share_y=args.share_y,
          ).show_all(show_mtl_sigmas=args.show_mtl_sigmas)


if __name__ == '__main__':
    main()
